# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""I/O for wave files"""
from pathlib import Path

import numpy as np

from .._data_obj import NDVar, Scalar, UTS
from .._types import PathArg
from .._utils import ui


FILETYPES = [("WAV files", "*.wav")]


def load_wav(
        filename: PathArg = None,
        name: str = None,
        backend: str = 'wave',
) -> NDVar:
    """Load a wav file as NDVar

    Parameters
    ----------
    filename
        Filename of the wav file. If not filename is specified, a file dialog is
        shown to select one.
    name
        NDVar name (default is the file name).
    backend : 'wave' | 'scipy'
        Whether to read the file using the builtin :mod:`wave` module or through
        :mod:`scipy.io.wavfile`.

    Returns
    -------
    wav
        NDVar with the wav file's data. If the file contains a single channel,
        the NDVar dimensions are ``(time,)``; if it contains several channels,
        they are ``(time, channel)``. ``wav.info`` contains entries for
        ``filename`` and ``samplingrate``.

    Notes
    -----
    Uses :mod:`scipy.io.wavfile`.
    """
    if filename is None:
        path = ui.ask_file("Load WAV File", "Select WAV file to load as NDVar", FILETYPES)
        if not path:
            return
    else:
        path = Path(filename)
        if not path.suffix and not path.exists():
            path = path.with_suffix('.wav')

    if backend == 'wave':
        import wave
        with wave.open(str(path), 'rb') as fp:
            n_channels = fp.getnchannels()
            n_frames = fp.getnframes()
            n_bytes = fp.getsampwidth()
            srate = fp.getframerate()
            data = fp.readframes(n_frames)
        data = np.frombuffer(data, f'<i{n_bytes}')
        if n_channels > 1:
            data = data.reshape((-1, n_channels))
    elif backend == 'scipy':
        from scipy.io import wavfile
        srate, data = wavfile.read(path)
    else:
        raise ValueError(f"backend={backend!r}")

    time = UTS(0, 1. / srate, data.shape[0])
    if name is None:
        name = path.name
    info = {'filename': str(path), 'samplingrate': srate}
    if data.ndim == 1:
        return NDVar(data, (time,), name, info)
    elif data.ndim == 2:
        chan = Scalar('channel', np.arange(data.shape[1]))
        return NDVar(data, (time, chan), name, info)
    else:
        raise NotImplementedError(f"Data with {data.ndim} dimensions")


def save_wav(ndvar, filename=None, toint=False):
    """Save an NDVar as wav file

    Parameters
    ----------
    ndvar : NDVar (time,)
        Sound data. Values should either be floating point numbers between -1
        and +1, or 16 bit integers.
    filename : path-like
        Where to save. If unspecified a file dialog will ask for the location.
    toint : bool
        Convert floating point data to 16 bit integer (default False).

    Notes
    -----
    Uses :mod:`scipy.io.wavfile`.
    """
    from scipy.io import wavfile

    data = ndvar.get_data('time')
    if toint and data.dtype != np.int16:
        above = data >= 2**15
        below = data < -2**15
        if np.any(above) or np.any(below):
            n = np.sum(above) + np.sum(below)
            print("WARNING: clipping %i samples" % n)
            data[above] = 2**15 - 1
            data[below] = -2**15

        data = data.astype(np.int16)
    elif data.dtype.kind != 'i' and (data.max() > 1. or data.min() < -1.):
        raise ValueError("Floating point data should be in range [-1, 1]. Set "
                         "toint=True to save as 16 bit integer data.")

    if filename is None:
        msg = "Save %s..." % ndvar.name
        filename = ui.ask_saveas(msg, msg, FILETYPES)
    srate = int(round(1. / ndvar.time.tstep))
    wavfile.write(filename, srate, data)



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""Tools for loading data form eyelink edf files."""
from glob import glob
import os
import re
import shutil
import subprocess
import tempfile

import numpy as np

from .._data_obj import Dataset, Datalist, Var
from .._utils import ui
from .._utils.subp import get_bin

__all__ = ('Edf', 'read_edf', 'read_edf_events', 'read_edf_samples')


class Edf:
    """Eyelink .edf file reader.

    Reads an eyelink .edf file and extracts epoch acceptability
    based on contamination with ocular artifacts (saccades and blinks).

    Parameters
    ----------
    path : str(path) | None
        Path to the .edf file. The If path contains '*', the files matching
        the pattern are concatenated. If None, a file-open dialogue will be
        displayed.

    Notes
    -----
    An edf file reader is initialized with the path to the corresponding
    file::

        >>> path = '/path/to/edf.edf'
        >>> edf = load.eyelink.Edf(path)

    There are several ways of retrieving trial acceptability values (see the
    used methods' documentation for more information):


    1. **For all triggers:**
       Acceptability for all triggers can be added to a Dataset with a single
       command if the Dataset contains the same events as the edf file::

          >>> edf.mark_all(ds, ...)


    2. **For a subset of triggers:**
       Often it is more efficient to compute acceptability only for a subset of
       the triggers contained in the edf file. For those cases, the trigger time
       should first be added to the complete Dataset with::

           >>> edf.add_t_to(ds)

       Now, the Dataset can be decimated::

           >>> ds = ds.sub(...)

       and acceptability can be added for the subset::

           >>> edf.mark(ds, ...)

    3. **Customized:**
       For a more customized use, all triggers from the edf can be retrieved using::

           >>> ds_edf = edf.get_triggers()

       The ``ds_edf['t_edf']`` time variable can be used to add trigger time values to
       arbitrary events, which can then in turn be used with::

           >>> edf.mark(...)

    """

    def __init__(self, path=None, samples=False):
        if path is None:
            path = ui.ask_file("Load an eyelink .edf file", "Pick the edf file",
                               [('eyelink data format (*.edf)', '*.edf')])

        # find all paths from which to read
        self.path = path
        if '*' in path:
            self.paths = glob(path)
        else:
            self.paths = [path]

        triggers = []
        artifacts = []
        pos = []
        for path in self.paths:
            edf_asc = read_edf(path)
            triggers += find_edf_triggers(edf_asc)
            artifacts += find_edf_artifacts(edf_asc)
            if samples:
                pos += find_edf_pos(edf_asc)

        dtype = [('T', np.uint32), ('Id', np.uint8)]
        self.triggers = np.array(triggers, dtype=dtype)

        dtype = np.dtype([('event', np.str_, 6), ('start', np.uint32), ('stop', np.uint32)])
        self.artifacts = np.array(artifacts, dtype=dtype)

        self.has_samples = bool(samples)
        if samples:
            self.time = np.array([item[0] for item in pos], dtype=np.uint32)
            self.xpos = np.array([item[1] for item in pos], dtype=np.float16)
            self.ypos = np.array([item[2] for item in pos], dtype=np.float16)
            self.pdia = np.array([item[3] for item in pos], dtype=np.float16)

    def __getstate__(self):
        state = {'path': self.path, 'paths': self.paths,
                 'triggers': self.triggers, 'artifacts': self.artifacts,
                 'has_samples': self.has_samples}
        if self.has_samples:
            state.update(time=self.time, xpos=self.xpos, ypos=self.ypos,
                         pdia=self.pdia)
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)

    def __repr__(self):
        return "Edf(%r)" % self.path

    def assert_trigger_match(self, ds=None, trigger='trigger'):
        """
        Make sure the Edf and another event list describe the same events.

        Raises an error if the triggers in ``trigger`` do not match the
        triggers in the Edf file(s).

        Parameters
        ----------
        ds : Dataset
            Dataset with events.
        trigger : str | array
            If ``ds`` is a Dataset, ``trigger`` should be a string naming the
            variable in `ds` containing the trigger values. If ``ds`` is
            ``None``, ``trigger`` should be a sequence of event triggers.
        """
        if isinstance(trigger, str):
            trigger = ds[trigger]

        edf_trigger = self.triggers['Id']
        if len(trigger) != len(edf_trigger):
            lens = (len(trigger), len(edf_trigger))
            mm = min(lens)
            for i in range(mm):
                if trigger[i] != edf_trigger[i]:
                    mm = i
                    break

            args = (getattr(ds, 'name', 'None'), self.path) + lens + (mm,)
            err = ("Dataset %r containes different number of events from edf "
                   "file %r (%i vs %i); first mismatch at %i." % args)
            raise ValueError(err)

        check = (trigger == edf_trigger)
        if not all(check):
            err = "Event ID mismatch: %s" % np.where(check == False)[0]
            raise ValueError(err)

    def add_t_to(self, ds, trigger='trigger', t_edf='t_edf'):
        """Add EDF trigger times as a variable to Dataset ds.

        These trigger times can then be used for Edf.add_by_T(ds) after ds hads
        been decimated.

        Parameters
        ----------
        ds : Dataset
            The Dataset to which the variable is added
        trigger : str | Var | None
            variable (or its name in the Dataset) containing event trigger
            values. Values in this variable are checked against the events in
            the EDF file, and an error is raised if there is a mismatch. This
            test can be skipped by setting trigger=None.
        t_edf : str
            Name for the target variable holding the edf trigger times.

        """
        if trigger:
            self.assert_trigger_match(ds=ds, trigger=trigger)
            if isinstance(trigger, str):
                trigger = ds[trigger]

        ds[t_edf] = Var(self.triggers['T'])

    def filter(self, ds, tstart=-0.1, tstop=0.6, use=['ESACC', 'EBLINK'],
               T='t_edf'):
        """Remove bad events from ``ds``

        Return a copy of the Dataset ``ds`` with all bad events removed. A
        Dataset containing all the bad events is stored in
        ``ds.info['rejected']``.

        Parameters
        ----------
        ds : Dataset
            The Dataset that is to be filtered
        tstart : scalar
            Start of the time window in which to look for artifacts
        tstop : scalar
            End of the time window in which to look for artifacts
        use : list of str
            List of events types which are to be treated as artifacts (possible
            are 'ESACC' and 'EBLINK')
        T : str | Var
            Variable describing edf-relative timing for the events in ``ds``.
            Usually this is a string key for a variable in ``ds``.

        """
        if isinstance(T, str):
            T = ds[T]
        accept = self.get_accept(T, tstart=tstart, tstop=tstop, use=use)
        accepted = ds.sub(accept)
        rejected = ds.sub(accept == False)
        accepted.info['rejected'] = rejected
        return accepted

    def get_accept(self, T=None, tstart=-0.1, tstop=0.6, use=['ESACC', 'EBLINK']):
        """Find good epochs

        Return a boolean Var indicating for each epoch whether it should be
        accepted or not based on ocular artifacts in the edf file.

        Parameters
        ----------
        T : array-like | None
            List of time points (in the edf file's time coordinates). If None,
            the edf's trigger events are used.
        tstart : scalar
            start of the epoch relative to the event (in seconds)
        tstop : scalar
            end of the epoch relative to the even (in seconds)

        """
        if T is None:
            T = self.triggers['T']

        # conert to ms
        start = int(tstart * 1000)
        stop = int(tstop * 1000)

        self._debug = []

        # get data for triggers
        N = len(T)
        accept = np.empty(N, np.bool_)

        x = tuple(self.artifacts['event'] == name for name in use)
        idx = np.any(x, axis=0)
        artifacts = self.artifacts[idx]

        for i, t in enumerate(T):
            starts_before_tstop = artifacts['start'] < t + stop
            stops_after_tstart = artifacts['stop'] > t + start
            overlap = np.all((starts_before_tstop, stops_after_tstart), axis=0)
            accept[i] = not np.any(overlap)

            self._debug.append(overlap)

        return accept

    def get_t(self, name='t_edf'):
        "Retrieve all trigger times in the Dataset"
        return Var(self.triggers['T'], name=name)

    def get_triggers(self, trigger='trigger', t_edf='t_edf'):
        """
        Return a Dataset with triggers and corresponding edf time values

        Parameters
        ----------
        trigger : str
            Name for the trigger value variable.
        t_edf : str
            Name for the variable containing edf event times.

        Returns
        -------
        ds : Dataset
            Dataset with triggers and corresponding edf times.
        """
        ds = Dataset(info={'edf': self})
        ds[trigger] = Var(self.triggers['Id'])
        ds[t_edf] = self.get_t(name=t_edf)
        return ds

    def mark(self, ds, tstart=-0.1, tstop=0.6, good=None, bad=False,
             use=['ESACC', 'EBLINK'], T='t_edf', target='accept'):
        """Mark events in ``ds`` as acceptable or not.

        ``ds`` needs to contain edf trigger times in a variable whose name is
        specified by the ``T`` argument.

        Parameters
        ----------
        ds : Dataset
            Dataset that contains the data to work with.
        tstart : scalar
            start of the time window relevant for rejection.
        tstop : scalar
            stop of the time window relevant for rejection.
        good : bool | None
            vale assigned to epochs that should be retained based on
            the eye-tracker data.
        bad : bool | None
            value that is assigned to epochs that should be rejected
            based on the eye-tracker data.
        use : list of str
            Artifact categories to include
        T : Var
            variable providing the trigger time values
        target : Var
            variable to which the good/bad values are assigned (if it does not
            exist, a new variable will be created with all values True
            initially)

        """
        if isinstance(target, str):
            if target in ds:
                target = ds[target]
            else:
                ds[target] = target = Var(np.ones(ds.n_cases, dtype=bool))

        if isinstance(T, str):
            T = ds[T]

        accept = self.get_accept(T, tstart=tstart, tstop=tstop, use=use)
        if good is not None:
            target[accept] = good
        if bad is not None:
            target[np.invert(accept)] = bad

    def mark_all(self, ds, tstart=-0.1, tstop=0.6, good=None, bad=False,
                 use=['ESACC', 'EBLINK'],
                 trigger='trigger', target='accept'):
        """
        Mark epochs in ``ds`` based on blinks and saccades

        Mark each epoch in ``ds`` for acceptability based on overlap with
        blinks and saccades. ds needs to contain the same number of triggers
        as the edf file. For adding acceptability to a decimated ds, use
        Edf.add_t_to(ds, ...) and then Edf.mark(ds, ...).

        Parameters
        ----------
        ds : Dataset
            Dataset that contains the data to work with.
        tstart : scalar
            start of the time window relevant for rejection.
        tstop : scalar
            stop of the time window relevant for rejection.
        good :
            vale assigned to epochs that should be retained based on
            the eye-tracker data.
        bad :
            value that is assigned to epochs that should be rejected
            based on the eye-tracker data.
        use : list of str
            Artifact categories to include
        trigger : Var | None
            variable containing trigger values for asserting that the Dataset
            contains the same triggers as the edf file(s). The test can be
            skipped by setting ``trigger=None``.
        target : Var
            variable to which the good/bad values are assigned (if it does not
            exist, a new variable will be created with all values True
            initially)
        """
        if trigger:
            self.assert_trigger_match(ds=ds, trigger=trigger)
            if isinstance(trigger, str):
                trigger = ds[trigger]

        T = self.get_t()
        self.mark(ds, tstart=tstart, tstop=tstop, good=good, bad=bad, use=use,
                  T=T, target=target)


def events(path, samples=False, ds=None, trigger='trigger', t_edf='t_edf'):
    """Read events from an edf file

    Parameters
    ----------
    path : str
        Filename.
    samples : bool
        Read continuous eye position data as well as events. This is needed to
        extract eye position data later.
    ds : Dataset
        Existing Dataset to which the edf-triggers should be added. If the
        Dataset contains a variable called 'trigger' whose content does not
        match the edf triggers, a ValueError is raised. ds is always modified
        in place, but returned for consistency.
    trigger : str
        Name of the trigger variable.
    t_edf : str
        Name of the edf time variable.

    Returns
    -------
    ds : Dataset
        Dataset with events form the edf file (if the ``ds`` input argument is
        provided, the ds that is returned is the same object as the input ds).
    """
    edf = Edf(path, samples=samples)
    if ds is None:
        ds = edf.get_triggers(trigger, t_edf)
    else:
        if 'edf' in ds.info:
            raise ValueError("ds.info already contains 'edf' entry.")
        if trigger in ds:
            edf.assert_trigger_match(ds)
        else:
            ds[trigger] = Var(edf.triggers['Id'])
        ds.info['edf'] = edf
        edf.add_t_to(ds, trigger, t_edf)

    return ds


def artifact_epochs(ds, tmin=-0.2, tmax=0.6, crop=True, esacc='sacc',
                    eblink='blink', t_edf='t_edf'):
    """Find blink and saccade artifact information by event

    Parameters
    ----------
    ds : Dataset
        Dataset with 'edf' entry in its info dictionary (usually a Dataset
        returned by ``load.eyelink.events()``)
    tmin, tmax : scalar
        Relative start and end points of the epoch (in seconds).
    crop : bool
        Crop events to epoch beginning and end (i.e., if an artifact starts
        before the epoch, set its start to the first sample in the epoch).
    esacc, eblink : None | str
        Name for the variable containing the corresponding information. If
        None, the corresponding variable is not added.
    t_edf : str
        Name of the ds variable containing edf times.

    Returns
    -------
    sacc : Datalist
        Saccade periods.
    blink : Datalist
        Blink periods.
    """
    edf = ds.info['edf']
    start = edf.artifacts['start']
    stop = edf.artifacts['stop']
    is_blink = (edf.artifacts['event'] == 'EBLINK')
    is_sacc = (edf.artifacts['event'] == 'ESACC')

    # edf times are in ms; convert them to s:
    dtype = np.dtype([('event', np.str_, 6), ('start', np.float64), ('stop', np.float64)])
    artifacts_s = edf.artifacts.astype(dtype)
    artifacts_s['start'] /= 1000.
    artifacts_s['stop'] /= 1000.

    sacc = Datalist(name=esacc) if esacc else None
    blink = Datalist(name=eblink) if eblink else None
    for t in ds[t_edf]:
        t_s = t / 1000.
        epoch_idx = np.logical_and(stop > t + tmin, start < t + tmax)
        if esacc:
            idx = np.logical_and(epoch_idx, is_sacc)
            epoch = artifacts_s[idx]
            epoch['start'] -= t_s
            epoch['stop'] -= t_s
            if crop and len(epoch):
                if epoch['start'][0] < tmin:
                    epoch['start'][0] = tmin
                if epoch['stop'][-1] > tmax:
                    epoch['stop'][-1] = tmax
            sacc.append(epoch)
        if eblink:
            idx = np.logical_and(epoch_idx, is_blink)
            epoch = artifacts_s[idx]
            epoch['start'] -= t_s
            epoch['stop'] -= t_s
            if crop and len(epoch):
                if epoch['start'][0] < tmin:
                    epoch['start'][0] = tmin
                if epoch['stop'][-1] > tmax:
                    epoch['stop'][-1] = tmax
            blink.append(epoch)

    return sacc, blink



def add_artifact_epochs(ds, tmin=-0.2, tmax=0.6, crop=True, esacc='sacc',
                        eblink='blink', t_edf='t_edf'):
    """Add a Datalist containing artifact information by event

    Parameters
    ----------
    ds : Dataset
        Dataset with 'edf' entry in its info dictionary (usually a Dataset
        returned by ``load.eyelink.events()``)
    tmin, tmax : scalar
        Relative start and end points of the epoch (in seconds).
    crop : bool
        Crop events to epoch beginning and end (i.e., if an artifact starts
        before the epoch, set its start to the first sample in the epoch).
    esacc, eblink : None | str
        Name for the variable containing the corresponding information. If
        None, the corresponding variable is not added.
    t_edf : str
        Name of the ds variable containing edf times.

    Returns
    -------
    ds : Dataset
        Return the input Dataset for consistency with similar functions; the
        Dataset is modified in place.
    """
    sacc, blink = artifact_epochs(ds, tmin, tmax, crop, esacc, eblink, t_edf)

    if esacc:
        ds.add(sacc)
    if eblink:
        ds.add(blink)
    return ds


def read_edf(fname, what='events'):
    """
    Read the content of an edf file as text using edf2asc

    Convert an "eyelink data format" (.edf) file to a temporary directory
    and parse its content.

    Parameters
    ----------
    fname : str
        Filename.
    what : 'all' | 'events' | 'samples'
        What type of information to read
    """
    if not os.path.isfile(fname):
        err = "%r is not a file." % fname
        raise ValueError(err)

    temp_dir = tempfile.mkdtemp()

    # edf2asc does not seem to handle spaces in filenames?
    if ' ' in fname:
        dst = os.path.join(temp_dir, os.path.basename(fname).replace(' ', '_'))
#         shutil.copy(fname, dst)
        os.symlink(fname, dst)
        fname = dst

    # construct the conversion command
    cmd = [get_bin('edfapi', 'edf2asc'),  # options in Manual p. 106
           '-t', ]  # use only tabs as delimiters
    if what == 'events':
        cmd.append('-e')  # outputs event data only
    elif what == 'samples':
        cmd.append('-s')  # outputs sample data only
    elif what == 'all':
        raise NotImplementedError()
    else:
        raise ValueError("what must be 'events' or 'samples', not %r" % what)

    cmd.extend(('-nst',  # blocks output of start events
                '-p', temp_dir,  # writes output with same name to <path> directory
                fname))

    # run the subprocess
    p = subprocess.Popen(cmd)
    stdout, stderr = p.communicate()
    # Don't check return code because it always return 255

    # find asc file
    name, _ = os.path.splitext(os.path.basename(fname))
    ascname = os.path.extsep.join((name, 'asc'))
    asc_path = os.path.join(temp_dir, ascname)
    if not os.path.exists(asc_path):
        print("======\nstdout\n======\n%s" % stdout)
        print("======\nstderr\n======\n%s" % stderr)
        raise subprocess.CalledProcessError(p.returncode, cmd, (stdout, stderr))
    with open(asc_path) as asc_file:
        asc_str = asc_file.read()

    # clean
    shutil.rmtree(temp_dir)
    return asc_str


def find_edf_triggers(asc_str):
    """Find artifacts in an edf asci representation

    Parameters
    ----------
    asc_str : str
        String with edf asci represenation as returned by read_edf.
    """
    re_trigger = re.compile(r'\bMSG\t(\d+)\tMEG Trigger: (\d+)')
    triggers = re_trigger.findall(asc_str)
    return triggers


def find_edf_artifacts(asc_str, kind='EBLINK|ESACC'):
    """Find artifacts in an edf asci representation

    Parameters
    ----------
    asc_str : str
        String with edf asci represenation as returned by read_edf.
    kind : 'EBLINK|ESACC' | 'EBLINK' | 'ESACC'
        Kind of artifact to search.
    """
    for kind_part in kind.split('|'):
        if kind_part not in ['EBLINK', 'ESACC']:
            raise ValueError("invalid kind parameter: %r" % kind)
    re_artifact = re.compile(r'\b(%s)\t[LR]\t(\d+)\t(\d+)' % kind)
    artifacts = re_artifact.findall(asc_str)
    return artifacts


def find_edf_pos(asc_str):
    """Find position values in an edf asci representation

    Parameters
    ----------
    asc_str : str
        String with edf asci represenation as returned by read_edf.
    """
    re_pos = re.compile(r'\b(\d+)\t(\d+\.\d*)\t(\d+\.\d*)\t(\d+\.\d*)')
    pos = re_pos.findall(asc_str)
    return pos



'''
Created on Dec 21, 2010

Mpl examples:
http://matplotlib.sourceforge.net/examples/user_interfaces/index.html


'''
from logging import getLogger
import tempfile

import numpy as np
from matplotlib.backends.backend_wx import NavigationToolbar2Wx
from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg
from matplotlib.backend_bases import FigureCanvasBase, MouseEvent
from matplotlib.figure import Figure
import wx

from .app import get_app
from .frame import EelbrainFrame, EelbrainDialog
from .help import show_help_txt
from .utils import FloatValidator, Icon
from . import ID


class FigureCanvasPanel(FigureCanvasWxAgg):
    """wx.Panel with a matplotlib figure

    Notes
    -----
    Subclass of mpl's Canvas to allow for more interaction with Eelbrain (such
    as copying the contents to the clipboard).
    """
    def __init__(self, parent, *args, **kwargs):
        """wx.Panel with a matplotlib figure

        Parameters
        ----------
        figsize : tuple
            Figure dimensions (width, height) in inches
        dpi : int
            Dots per inch.
        facecolor : mpl color
            The figure patch facecolor; defaults to rc ``figure.facecolor``
        edgecolor : mpl color
            The figure patch edge color; defaults to rc ``figure.edgecolor``
        linewidth : scalar
            The figure patch edge linewidth; the default linewidth of the frame
        frameon : bool
            If ``False``, suppress drawing the figure frame
        subplotpars :
            A :class:`SubplotParams` instance, defaults to rc
        tight_layout : bool | dict
            If ``False`` use ``subplotpars``; if ``True`` adjust subplot
            parameters using :meth:`tight_layout` with default padding.
            When providing a dict containing the keys `pad`, `w_pad`, `h_pad`
            and `rect`, the default :meth:`tight_layout` paddings will be
            overridden. Defaults to rc ``figure.autolayout``.
        """
        self.figure = Figure(*args, **kwargs)
        FigureCanvasWxAgg.__init__(self, parent, wx.ID_ANY, self.figure)
        self.Bind(wx.EVT_ENTER_WINDOW, self.ChangeCursor)
        self._background = None

    def _onKeyDown(self, evt):
        # Override to avoid system chime
        FigureCanvasBase.key_press_event(self, self._get_key(evt), guiEvent=evt)

    def _onKeyUp(self, evt):
        # Override to avoid system chime
        FigureCanvasBase.key_release_event(self, self._get_key(evt), guiEvent=evt)

    def CanCopy(self):
        return True

    def bufferHasChanged(self):
        return True

    def ChangeCursor(self, event):
        "http://matplotlib.sourceforge.net/examples/user_interfaces/wxcursor_demo.html"
        self.SetCursor(wx.Cursor(wx.CURSOR_CROSS))

    def Copy(self):
        # By default, copy PDF
        if not wx.TheClipboard.Open():
            getLogger('eelbrain').debug("Failed to open clipboard")
            return
        try:
            path = tempfile.mktemp('.pdf')
            self.figure.savefig(path)
            # copy path
            do = wx.FileDataObject()
            do.AddFile(path)
            wx.TheClipboard.SetData(do)
        finally:
            wx.TheClipboard.Close()
            wx.TheClipboard.Flush()

    @staticmethod
    def CanCopyPNG():
        return True

    def CopyAsPNG(self):
        self.Copy_to_Clipboard()

    def _to_matplotlib_event(self, event, name='wx-event', button=None, key=None):
        """Convert wxPython event to Matplotlib event

        - Sets axes and position but ignores source
        - cf. matplotlib.backends.backend_wx._FigureCanvasWxBase
        """
        x = event.GetX()
        y = self.figure.bbox.height - event.GetY()
        return MouseEvent(name, self.figure.canvas, x, y, button, key, guiEvent=event)

    def redraw(self, axes=set(), artists=()):
        # FIXME:  redraw artist instead of whole axes
        if artists:
            axes.update(artist.axes for artist in artists)
        elif not axes:
            return
        elif self._background is None:
            raise RuntimeError("Background not captured")

        self.restore_region(self._background)
        for ax in axes:
            ax.draw_artist(ax)
            self.blit(ax.get_window_extent())
        # for artist in artists:
        #     artist.axes.draw_artist(artist.axes)
        #     self.blit(artist.axes.get_window_extent())

    def store_canvas(self):
        self._background = self.copy_from_bbox(self.figure.bbox)


class CanvasFrame(EelbrainFrame):
    # after:
    # http://matplotlib.sourceforge.net/examples/user_interfaces/embedding_in_wx2.html
    _plot_name = "CanvasFrame"
    _allow_user_set_title = True

    def __init__(self, parent=None, title="Matplotlib Frame", pos=wx.DefaultPosition, eelfigure=None, statusbar=True, toolbar=True, mpl_toolbar=False, **kwargs):
        EelbrainFrame.__init__(self, parent, -1, title, pos)
        self._pos_arg = pos

        # set up the canvas
        self.sizer = sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer)
        self.canvas = FigureCanvasPanel(self, **kwargs)
        sizer.Add(self.canvas, 1, wx.EXPAND)
        self.figure = self.canvas.figure

        if statusbar:
            self.CreateStatusBar()

        if toolbar:
            tb = self.CreateToolBar(wx.TB_HORIZONTAL)
            tb.SetToolBitmapSize(size=(32, 32))
            self.FillToolBar(tb, eelfigure)
            tb.Realize()

        if mpl_toolbar:
            self.add_mpl_toolbar()

        self.Fit()
        self._eelfigure = eelfigure
        self.Bind(wx.EVT_CLOSE, self.OnClose)

    def store_canvas(self):
        self.canvas.store_canvas()

    def FillToolBar(self, tb, eelfigure):
        "Subclass should call this after adding their own items"
        if hasattr(self.Parent, 'attach'):
            tb.AddTool(ID.ATTACH, "Attach", Icon("actions/attach"))
            self.Bind(wx.EVT_TOOL, self.OnAttach, id=ID.ATTACH)

        tb.AddTool(wx.ID_SAVE, "Save", Icon("tango/actions/document-save"))
        self.Bind(wx.EVT_TOOL, self.OnSaveAs, id=wx.ID_SAVE)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUISave, id=wx.ID_SAVE)

        # intermediate, custom part
        if eelfigure is None:
            self._fill_toolbar(tb)
        else:
            eelfigure._fill_toolbar(tb)
            self._plot_name = eelfigure.__class__.__name__
            self.__doc__ = eelfigure.__doc__

        # right-most part
        tb.AddStretchableSpace()
        tb.AddTool(ID.ATTACH, "Attach", Icon("actions/attach"))
        self.Bind(wx.EVT_TOOL, self.OnAttach, id=ID.ATTACH)
        if self.__doc__:
            tb.AddTool(wx.ID_HELP, 'Help', Icon("tango/apps/help-browser"))
            self.Bind(wx.EVT_TOOL, self.OnHelp, id=wx.ID_HELP)

    def _fill_toolbar(self, tb):
        pass

    def Show(self, show=True):
        if self._pos_arg == wx.DefaultPosition:
            rect = self.GetRect()
            display_w, display_h = wx.DisplaySize()
            # print(self.Rect.Right)
            dx = -rect.Left if rect.Right > display_w else 0
            dy = -rect.Top + 22 if rect.Bottom > display_h else 0
            if dx or dy:
                rect.Offset(dx, dy)
                self.SetRect(rect)
        EelbrainFrame.Show(self, show)

    def add_mpl_toolbar(self):
        self.toolbar = NavigationToolbar2Wx(self.canvas)
        self.toolbar.Realize()
        if 0:  # wx.Platform == '__WXMAC__':
            # Mac platform (OSX 10.3, MacPython) does not seem to cope with
            # having a toolbar in a sizer. This work-around gets the buttons
            # back, but at the expense of having the toolbar at the top
            self.SetToolBar(self.toolbar)
        else:
            # On Windows platform, default window size is incorrect, so set
            # toolbar width to figure width.
            tw, th = self.toolbar.GetSizeTuple()
            fw, fh = self.canvas.GetSizeTuple()
            # By adding toolbar in sizer, we are able to put it at the bottom
            # of the frame - so appearance is closer to GTK version.
            # As noted above, doesn't work for Mac.
            self.toolbar.SetSize(wx.Size(fw, th))
            self.Sizer.Add(self.toolbar, 0, wx.LEFT | wx.EXPAND)
        # update the axes menu on the toolbar
        self.toolbar.update()

    def OnAttach(self, event):
        get_app().Attach(self._eelfigure, "%s plot" % self._plot_name, 'p', self)

    def OnClose(self, event):
        # remove circular reference
        if getattr(self, '_eelfigure', None):
            self._eelfigure._frame = None
            del self._eelfigure
        event.Skip()

    def OnDrawCrosshairs(self, event):
        self._eelfigure.draw_crosshairs(event.IsChecked())

    def OnHelp(self, event):
        show_help_txt(self.__doc__, self, self._plot_name)

    def OnSave(self, event):
        self.OnSaveAs(event)

    def OnSaveAs(self, event):
        default_file = '%s.pdf' % self.GetTitle().replace(': ', ' - ')
        dlg = wx.FileDialog(self, "If no file type is selected below, it is "
                                  "inferred from the extension.",
                            defaultFile=default_file,
                            wildcard="Any (*.*)|*.*|PDF (*.pdf)|*.pdf|PNG (*.png)|*.png",
                            style=wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)
        if dlg.ShowModal() == wx.ID_OK:
            self.figure.savefig(dlg.GetPath())
        dlg.Destroy()

    def OnSetTime(self, event):
        dlg = SetTimeDialog(self, self._eelfigure.get_time())
        if dlg.ShowModal() == wx.ID_OK:
            self._eelfigure.set_time(dlg.time)
        dlg.Destroy()

    def OnSetVLim(self, event):
        if self._eelfigure._can_set_vlim:
            vlim = self._eelfigure.get_vlim()
            ylim = None
        else:
            vlim = None
            ylim = self._eelfigure.get_ylim() if self._eelfigure._can_set_ylim else None
        xlim = self._eelfigure.get_xlim() if self._eelfigure._can_set_xlim else None
        dlg = AxisLimitsDialog(vlim, ylim, xlim, self)
        if dlg.ShowModal() == wx.ID_OK:
            if vlim is not None:
                self._eelfigure.set_vlim(*dlg.vlim)
            if ylim is not None:
                self._eelfigure.set_ylim(*dlg.ylim)
            if xlim is not None:
                self._eelfigure.set_xlim(*dlg.xlim)
        dlg.Destroy()

    def OnUpdateUIDrawCrosshairs(self, event):
        event.Enable(True)
        event.Check(self._eelfigure._draw_crosshairs)

    def OnUpdateUISave(self, event):
        event.Enable(True)

    def OnUpdateUISaveAs(self, event):
        event.Enable(True)

    def OnUpdateUISetVLim(self, event):
        event.Enable(self._eelfigure._can_set_xlim or self._eelfigure._can_set_ylim)

    def OnUpdateUISetTime(self, event):
        event.Enable(self._eelfigure._can_set_time)


class TestCanvas(CanvasFrame):
    "This is a minimal CanvasFrame subclass"
    def __init__(self, effect=10, mpl_toolbar=True):
        CanvasFrame.__init__(self, title="Test MPL Frame", mpl_toolbar=mpl_toolbar)
        self.plot()
        self.Show(effect)

    def plot(self):
        self.axes = self.figure.add_subplot(111)
        t = np.arange(0.0, 3.0, 0.01)
        s = np.sin(2 * np.pi * t)
        self.axes.plot(t, s)


class LimitsValidator(wx.Validator):
    """Validator for axis limits

    Parameters
    ----------
    parent : wx Dialog
        Parent.
    attr : str
        Attribute on ``parent`` to operate on.
    single : 'symmetric' | 'offset'
        How to interpret input when only one value is provided;
        ``symmetric``: ``[-v, v]``
        ``offset``: ``[v, v + old_range]``
    """

    def __init__(self, parent, attr, single='symmetric'):
        wx.Validator.__init__(self)
        self.parent = parent
        self.attr = attr
        self.value = None
        self.single = single

    def Clone(self):
        return LimitsValidator(self.parent, self.attr, self.single)

    def Validate(self, parent):
        ctrl = self.GetWindow()
        value = ctrl.GetValue()
        if not value.strip():
            self.TransferToWindow()
            return False

        try:
            values = value.replace(',', ' ').split()
            if len(values) == 1:
                if self.single == 'symmetric':
                    upper = float(values[0])
                    lower = -upper
                elif self.single == 'offset':
                    lower = float(values[0])
                    old_lower, old_upper = getattr(self.parent, self.attr)
                    upper = lower + old_upper - old_lower
                else:
                    raise RuntimeError(f'single={self.single!r}')
            elif len(values) == 2:
                lower, upper = map(float, values)
            else:
                raise ValueError("Wrong number of values")
        except Exception as exception:
            msg = wx.MessageDialog(
                self.parent, str(exception), "Invalid Entry",
                wx.OK | wx.ICON_ERROR)
            msg.ShowModal()
            msg.Destroy()
            return False
        else:
            self.value = (lower, upper)
            return True

    def TransferToWindow(self):
        ctrl = self.GetWindow()
        ctrl.SetValue("%s %s" % getattr(self.parent, self.attr))
        ctrl.SelectAll()
        return True

    def TransferFromWindow(self):
        if self.value is None:
            return False
        else:
            setattr(self.parent, self.attr, self.value)
            return True


class AxisLimitsDialog(EelbrainDialog):
    """Dialog to set axis limits

    Parameters
    ----------
    ylim : None | tuple
        ``(bottom, top)`` if the y-limits can be adjusted, else ``None``.
    xlim : None | tuple
        ``(left, right)`` if the x-limits can be adjusted, else ``None``.
    ... :
        Wx-Python Dialog parameters.
    """

    def __init__(self, vlim, ylim, xlim, parent, *args, **kwargs):
        EelbrainDialog.__init__(self, parent, wx.ID_ANY, "Set Axis Limits",
                                *args, **kwargs)
        self.vlim = vlim
        self.ylim = ylim
        self.xlim = xlim

        # Layout
        mainsizer = wx.BoxSizer(wx.VERTICAL)

        sizer = wx.GridSizer(2, 5, 5)
        # v-limit
        if vlim is not None:
            sizer.Add(wx.StaticText(self, label="Value Limits:"))
            self.v_text = wx.TextCtrl(self, validator=LimitsValidator(self, 'vlim'))
            sizer.Add(self.v_text)
            self.v_text.SetFocus()
        else:
            self.v_text = None
        # y-axis
        if ylim is not None:
            sizer.Add(wx.StaticText(self, label="Y-Axis Limits:"))
            self.y_text = wx.TextCtrl(self, validator=LimitsValidator(self, 'ylim'))
            sizer.Add(self.y_text)
            self.y_text.SetFocus()
        else:
            self.y_text = None
        # x-axis
        if xlim is not None:
            sizer.Add(wx.StaticText(self, label="X-Axis Limits:"))
            self.x_text = wx.TextCtrl(self, validator=LimitsValidator(self, 'xlim', 'offset'))
            sizer.Add(self.x_text)
            if self.y_text is None:
                self.x_text.SetFocus()
        else:
            self.x_text = None

        # buttons
        button_sizer = wx.StdDialogButtonSizer()
        # ok
        btn = wx.Button(self, wx.ID_OK)
        btn.SetDefault()
        button_sizer.AddButton(btn)
        # cancel
        btn = wx.Button(self, wx.ID_CANCEL)
        button_sizer.AddButton(btn)
        button_sizer.Realize()

        # finalize
        mainsizer.Add(sizer, flag=wx.ALL, border=10)
        # mainsizer.AddSpacer(10)
        mainsizer.Add(button_sizer, flag=wx.ALL | wx.ALIGN_RIGHT, border=10)
        self.SetSizer(mainsizer)
        mainsizer.Fit(self)


class SetTimeDialog(EelbrainDialog):

    def __init__(self, parent, current_time):
        EelbrainDialog.__init__(self, parent, title="Set Time")
        self.time = current_time

        sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer.Add(wx.StaticText(self, label="New time:"))
        self.text_ctrl = wx.TextCtrl(self, validator=FloatValidator(self, 'time'))
        sizer.Add(self.text_ctrl)
        self.text_ctrl.SetFocus()

        # buttons
        button_sizer = wx.StdDialogButtonSizer()
        # ok
        btn = wx.Button(self, wx.ID_OK)
        btn.SetDefault()
        button_sizer.AddButton(btn)
        # cancel
        btn = wx.Button(self, wx.ID_CANCEL)
        button_sizer.AddButton(btn)
        button_sizer.Realize()

        # finalize
        mainsizer = wx.BoxSizer(wx.VERTICAL)
        mainsizer.Add(sizer, flag=wx.ALL, border=10)
        # mainsizer.AddSpacer(10)
        mainsizer.Add(button_sizer, flag=wx.ALL | wx.ALIGN_RIGHT, border=10)
        self.SetSizer(mainsizer)
        mainsizer.Fit(self)



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""NIST SPH audio-file format"""
from pathlib import Path

import numpy

from .._data_obj import NDVar, Scalar, UTS
from .._types import PathArg
from .._utils import ui


def load_sphere(
        filename: PathArg = None,
        name: str = None,
) -> NDVar:
    """Load NIST SPH audio-files (uses `sphfile <https://pypi.org/project/sphfile/>`_

    Parameters
    ----------
    filename
        Filename of the file. If not filename is specified, a file dialog is
        shown to select one.
    name
        NDVar name (default is the file name).
    """
    try:
        import sphfile
    except ImportError:
        raise ImportError("The sphfile library needs to be installed; run:\n $ pip install sphfile")

    if filename is None:
        path = ui.ask_file("Load NIST SPH audio", "Select NIST SPH audio file to load as NDVar", [("NIST SPH audio-file", "*.WAV")])
        if not path:
            return
    else:
        path = Path(filename)
        if not path.suffix and not path.exists():
            path = path.with_suffix('.WAV')

    sphere_file = sphfile.SPHFile(path)
    time = UTS(0, 1. / sphere_file.format['sample_rate'], sphere_file.format['sample_count'])
    if name is None:
        name = path.name
    info = {'filename': str(path), **sphere_file.format}
    if sphere_file.content.ndim == 1:
        return NDVar(sphere_file.content, (time,), name, info)
    elif sphere_file.content.ndim == 2:
        chan = Scalar('channel', numpy.arange(sphere_file.content.shape[1]))
        return NDVar(sphere_file.content, (time, chan), name, info)
    else:
        raise NotImplementedError(f"Data with {sphere_file.content.ndim} dimensions")



"""
Tools for loading data from text files.

.. autosummary::
   :toctree: generated

   tsv
   var
"""
import csv
import os
from pathlib import Path
import re
from numbers import Number
from typing import Sequence, Union

import numpy as np

from .._types import PathArg
from .._utils import ui
from .._utils.parse import FLOAT_NAN_PATTERN
from .. import _data_obj as _data

__all__ = ('tsv', 'var')


def to_num(v: str):
    if v.isdigit():
        return int(v)
    else:
        return float(v)


# could use csv module:  http://docs.python.org/3/library/csv.html
def tsv(
        path: PathArg = None,
        names: Union[Sequence[str], bool] = True,
        types: Union[str, dict] = None,
        delimiter: str = 'auto',
        skiprows: int = 0,
        start_tag: str = None,
        ignore_missing: bool = False,
        empty: Union[str, float] = None,
        random: Union[str, Sequence[str]] = None,
        **fmtparams,
):
    r"""Load a :class:`Dataset` from a text file.

    Parameters
    ----------
    path : str
        Path to the file (if omitted, use a system file dialog).
    names : Sequence of str | bool
        Column/variable names.

        * ``True`` (default): look for names on the first line of the file
        * ``['name1', ...]`` use these names
        * ``False``: use "v1", "v2", ...
    types : str | dict
        Column data types::

         - 'a': determine automatically
         - 'f': Factor
         - 'v': Var
         - 'b': boolean Var

        Specified either as string, with one type per columnc (e.g.
        ``'ffvva'``) or as ``{column_name: data_type}`` dictionary (e.g.
        ``{'participant': 'f'}``); unspecified columns default to ``'a'``.
    delimiter : str
        Value delimiting cells in the input file (default depends on ``path``:
        ``','`` if the extension is ``'.csv'``, otherwise ``'\t'``).
    skiprows : int
        Skip so many rows at the beginning of the file (for tsv files with
        headers). Column names are expected to come after the skipped rows.
        ``skiprows`` is applied after ``start_tag``.
    start_tag : str
        Alternative way to skip header rows. The table is assumed to start
        on the line following the last line in the file that starts with
        ``start_tag``.
    ignore_missing : bool
        Ignore rows with missing values (i.e., lines that contain fewer
        ``delimiter`` than the others; by default this raises an IOError). For
        rows with missing values, ``NaN`` is substituted for numerical and
        ``""`` for categorial variables.
    empty : number | 'nan'
        For numerical variables, substitute this value for empty entries. For
        example, if a column in a file contains ``['5', '3', '']``, this is read
        by default as ``Factor(['5', '3', ''])``.
        With ``empty=0``, it is read as ``Var([5, 3, 0])``.
        With ``empty='nan'``, it is read as ``Var([5, 3, nan])``.
    random : str | sequence of str
        Names of the columns that should be assigned as random factor.
    **fmtparams
        Further formatting parameters for :func:`csv.reader`. For example, a
        fixed-width column file can be loaded with ``skipinitialspace=True``
        (as long as there are no empty cells).
    """
    if path is None:
        path = ui.ask_file("Load TSV", "Select tsv file to import as Dataset")
        if not path:
            return
    path = Path(path)

    if isinstance(random, str):
        random = [random]
    elif random is None:
        random = []
    else:
        random = list(random)

    if delimiter is None:  # legacy option
        delimiter = ' '
        fmtparams['skipinitialspace'] = True
    elif delimiter == 'auto':
        suffix = path.suffix.lower()
        if suffix == '.csv':
            delimiter = ','
        else:
            delimiter = '\t'

    with open(path, newline='') as fid:
        reader = csv.reader(fid, delimiter=delimiter, **fmtparams)
        lines = list(reader)

    # find start position
    if start_tag:
        start = 0
        for i, line in enumerate(lines, 1):
            if line[0].startswith(start_tag):
                start = i
        if start:
            lines = lines[start:]
    if skiprows:
        lines = lines[skiprows:]

    # read / create names
    if names is True:
        names = lines.pop(0)
    elif names:
        names = list(names)

    row_lens = set(len(row) for row in lines)
    if not ignore_missing and len(row_lens) > 1:
        raise IOError("Not all rows have same number of entries. Set ignore_missing to True in order to ignore this error.")
    n_cols = max(row_lens)
    n_rows = len(lines)

    if names:
        n_names = len(names)
        if n_names == n_cols - 1:
            # R write.table saves unnamed column with row names
            name = "row"
            while name in names:
                name += '_'
            names.insert(0, name)
        elif n_names != n_cols:
            raise IOError(f"The number of names in the header ({n_names}) does not correspond to the number of columns in the table ({n_cols})")
    else:
        names = [f'v{i}' for i in range(n_cols)]

    # check random
    missing = [k for k in random if k not in names]
    if missing:
        raise ValueError(f"random={random} includes non-existent names: {', '.join(missing)}")

    # coerce types parameter
    if types is None:
        types_ = 'a' * n_cols
    elif isinstance(types, dict):
        types_ = [types.get(n, 'a') for n in names]
    elif isinstance(types, str):
        types_ = types
    elif isinstance(types, (list, tuple)):
        # backwards compatibility
        types_ = ['afv'[v] for v in types]
    else:
        raise TypeError(f'types={types!r}')

    # check types values
    if len(types_) != n_cols:
        raise ValueError(f'types={types!r}: {len(types)} values for file with {n_cols} columns')
    elif set(types_).difference('afvb'):
        invalid = ', '.join(map(repr, set(types_).difference('afvb')))
        raise ValueError(f'types={types!r}: invalid values {invalid}')

    # find quotes (imply type 1)
    # quotes = "'\""
    data = np.empty((n_rows, n_cols), object)
    for r, line in enumerate(lines):
        for c, v in enumerate(line):
            # for str_del in quotes:
            #     if len(v) > 0 and v[0] == str_del:
            #         v = v.strip(str_del)
            #         types_[c] = 'f'
            data[r, c] = v

    # convert values to data-objects
    float_pattern = re.compile(FLOAT_NAN_PATTERN)
    ds = _data.Dataset(name=os.path.basename(path))
    np_vars = {
        'NA': np.nan,
        'na': np.nan,
        'NaN': np.nan,
        'nan': np.nan,
        'NAN': np.nan,
        None: np.nan,
    }
    if empty is not None:
        if isinstance(empty, str):
            np_vars[''] = to_num(empty)
        elif isinstance(empty, Number):
            np_vars[''] = empty
        else:
            raise TypeError(f'empty={empty!r}')
    bool_dict = {'True': True, 'False': False, None: False}
    keys = {}
    for name, values, type_ in zip(names, data.T, types_):
        # infer type
        if type_ in 'fvb':
            pass
        elif all(v in bool_dict for v in values):
            type_ = 'b'
        elif empty is not None:
            if all(v in (None, '') or float_pattern.match(v) for v in values):
                type_ = 'v'
            else:
                type_ = 'f'
        elif all(v is None or float_pattern.match(v) for v in values):
            type_ = 'v'
        else:
            type_ = 'f'

        # substitute values
        if type_ == 'v':
            values = [np_vars[v] if v in np_vars else to_num(v) for v in values]
        elif type_ == 'b':
            values = [bool_dict[v] for v in values]

        # create data-object
        if type_ == 'f':
            f_random = name in random
            dob = _data.Factor(values, labels={None: ''}, name=name, random=f_random)
        elif name in random:
            raise ValueError(f"random={random}: {name} is not categorial")
        else:
            dob = _data.Var(values, name)
        key = _data.Dataset.as_key(name)
        keys[name] = key
        ds[key] = dob

    if any(k != v for k, v in keys.items()):
        ds.info['keys'] = keys

    return ds


def var(path: PathArg = None, name: str = None):
    """
    Load a :class:`Var` object from a text file by splitting at white-spaces.

    Parameters
    ----------
    path : str
        Source file. If not specified, a system file dialog is opened.
    name : str
        Name for the Var.
    """
    if path is None:
        path = ui.ask_file("Load Var File", "Select Var File")
        if not path:
            return

    FILE = open(path)
    lines = FILE.read().split()
    FILE.close()
    is_bool = all(line in ['True', 'False'] for line in lines)

    if is_bool:
        x = np.genfromtxt(path, dtype=bool)
    else:
        x = np.loadtxt(path)

    return _data.Var(x, name)



"""Some basic example datasets for testing."""
from distutils.version import LooseVersion
from itertools import product
import os
from pathlib import Path
import shutil
import string

import mne
from mne import minimum_norm as mn
import numpy as np

from .. import _info, load
from .._data_obj import Dataset, Factor, Var, NDVar, Case, Categorial, Scalar, Sensor, Space, UTS
from .._ndvar import concatenate, convolve
from .._types import PathArg


def permute(variables):
    """Create a Dataset from permuting variables.

    Parameters
    ----------
    variables : sequence
        Sequence of (name, values) tuples. For examples:
        ``(('A', ('a1', 'a2')), ('B', ('b1', 'b2')))``

    Examples
    --------
    >>> print(permute((('A', ('a1', 'a2')),('B', ('b1', 'b2')))))
    A    B
    -------
    a1   b1
    a1   b2
    a2   b1
    a2   b2
    """
    names = (v[0] for v in variables)
    cases = tuple(product(*(v[1] for v in variables)))
    return Dataset.from_caselist(names, cases)


def _get_continuous(
        n_samples: int = 100,
        ynd: bool = False,
        seed: int = 0,
        xn: int = 0,
) -> Dataset:
    """Generate continuous data for reverse correlation

    Parameters
    ----------
    n_samples
        Number of samples to simulate.
    ynd
        Include 3d ``y``.
    seed
        Seed for :func:`numpy.random.seed` (``None`` to skip seeding random
        state; default us 0).
    xn
        Number of rows in ``xn``.

    Returns
    -------
    data
        Dataset with simulated data:

         - ``x1``: random time series
         - ``x2``: two random time series
         - ``h1`` and ``h2``: Kernels corresponding to ``x1`` and ``x2``
         - ``y``: convolution of ``(x1 * h1) + (x2 * h2)``
    """
    random = np.random if seed is None else np.random.RandomState(seed)
    time = UTS(0, 0.1, n_samples)
    h_time = UTS(0, 0.1, 10)
    xdim = Scalar('xdim', [0, 1])

    x1 = NDVar(random.normal(0, 1, (n_samples,)), (time,), 'x1')
    h1 = NDVar(np.array([0, 0, 1, 3, 0, 0, 0, 0, 2, 3]), (h_time,), 'h1')

    x2 = NDVar(random.normal(0, 1, (2, n_samples,)), (xdim, time), 'x2')
    h2 = NDVar(np.array([[0, 0, 0, 0, 0, 0, -1, -3, 0, 0],
                         [0, 0, 2, 2, 0, 0, 0, 0, 0, 0]]),
               (xdim, h_time), 'h2')

    y = convolve(h1, x1)
    y += convolve(h2, x2)
    y.name = 'y'
    y.info = _info.for_eeg()
    out = {'y': y, 'x1': x1, 'h1': h1, 'x2': x2, 'h2': h2}
    if ynd:
        dim = Sensor([[-1, 0, 0], [1, 0, 0]], ['SL', 'SR'], 'TEST-2-CH')
        out['h1nd'] = h1 = concatenate([h1, -h1], dim, 'h1')
        out['h2nd'] = h2 = concatenate([h2, h2 * 0], dim, 'h2')
        out['ynd'] = convolve(h1, x1) + convolve(h2, x2)
    if xn:
        xn_dim = Scalar('xdim', range(xn))
        out['xn'] = NDVar(random.normal(0, 1, (xn, n_samples)), (xn_dim, time))
        out['hn'] = NDVar(random.normal(0, 1, (xn, 10)), (xn_dim, h_time))
        out['yn'] = convolve(out['hn'], out['xn'])
        out['yn'] += random.normal(0, 1, (n_samples,))
    return Dataset(out)


def get_loftus_masson_1994():
    "Dataset used for illustration purposes by Loftus and Masson (1994)"
    ds = Dataset()
    ds['subject'] = Factor(range(1, 11), tile=3, random=True)
    ds['exposure'] = Var([1, 2, 5], repeat=10)
    ds['n_recalled'] = Var([10, 6, 11, 22, 16, 15, 1, 12, 9, 8,
                            13, 8, 14, 23, 18, 17, 1, 15, 12, 9,
                            13, 8, 14, 25, 20, 17, 4, 17, 12, 12])
    return ds


def get_mne_epochs():
    """MNE-Python Epochs"""
    data_path = mne.datasets.sample.data_path()
    raw_path = os.path.join(data_path, 'MEG', 'sample',
                            'sample_audvis_raw.fif')
    events_path = os.path.join(data_path, 'MEG', 'sample',
                               'sample_audvis_raw-eve.fif')
    raw = mne.io.Raw(raw_path)
    events = mne.read_events(events_path)
    epochs = mne.Epochs(raw, events, 32, -0.1, 0.4, preload=True)
    return epochs


def get_mne_evoked(ndvar=False):
    """MNE-Python Evoked

    Parameters
    ----------
    ndvar : bool
        Convert to NDVar (default False).
    """
    data_path = mne.datasets.sample.data_path()
    evoked_path = os.path.join(data_path, 'MEG', 'sample',
                               'sample_audvis-ave.fif')
    evoked = mne.Evoked(evoked_path, "Left Auditory")
    if ndvar:
        return load.fiff.evoked_ndvar(evoked)
    else:
        return evoked


def get_mne_stc(ndvar=False, src='ico-5', subject='sample'):
    """MNE-Python SourceEstimate

    Parameters
    ----------
    ndvar : bool
        Convert to NDVar (default False; src="ico-4" is false, but it works as
        long as the source space is not accessed).
    src : 'ico-5' | 'vol-7' | 'oct-4
        Source space to use.

    Notes
    -----
    Source space only available for ``oct-4``, ``sample``.
    """
    data_path = Path(mne.datasets.testing.data_path())
    meg_sdir = data_path / 'MEG' / 'sample'
    subjects_dir = data_path / 'subjects'
    # scaled subject
    if subject == 'fsaverage_scaled':
        subject_dir = os.path.join(subjects_dir, subject)
        if not os.path.exists(subject_dir):
            mne.scale_mri('fsaverage', subject, .9, subjects_dir=subjects_dir, skip_fiducials=True, labels=False, annot=True)
        data_subject = 'fsaverage'
    else:
        data_subject = subject

    if src == 'vol-7':
        inv = mn.read_inverse_operator(str(meg_sdir / 'sample_audvis_trunc-meg-vol-7-meg-inv.fif'))
        evoked = mne.read_evokeds(str(meg_sdir / 'sample_audvis_trunc-ave.fif'), 'Left Auditory')
        stc = mn.apply_inverse(evoked, inv, method='MNE', pick_ori='vector')
        if data_subject == 'fsaverage':
            m = mne.compute_source_morph(stc, 'sample', data_subject, subjects_dir)
            stc = m.apply(stc)
            stc.subject = subject
        elif subject != 'sample':
            raise ValueError(f"subject={subject!r}")
        if ndvar:
            return load.fiff.stc_ndvar(stc, subject, 'vol-7', subjects_dir, 'MNE', sss_filename='{subject}-volume-7mm-src.fif')
        else:
            return stc
    elif src == 'oct-4':
        if subject != 'sample':
            raise ValueError(f"subject={subject!r}: source space only available for 'sample'")
        inv = mne.minimum_norm.read_inverse_operator(meg_sdir / 'sample_audvis_trunc-meg-eeg-oct-4-meg-inv.fif')
        evokeds = mne.read_evokeds(meg_sdir / 'sample_audvis_trunc-ave.fif')
        evoked = mne.combine_evoked([evokeds[i].apply_baseline() for i in [0, 1]], [1, 1])
        stc = mne.minimum_norm.apply_inverse(evoked, inv)
    elif src == 'ico-5':
        stc_path = meg_sdir / f'{data_subject}_audvis_trunc-meg'
        stc = mne.read_source_estimate(str(stc_path), subject)
    else:
        raise ValueError(f"src={src!r}")

    if ndvar:
        return load.fiff.stc_ndvar(stc, subject, src, subjects_dir)
    else:
        return stc


def _mne_source_space(subject, src_tag, subjects_dir):
    """Load mne source space

    Parameters
    ----------
    subject : str
        Subejct
    src_tag : str
        Spacing (e.g., 'ico-4').
    """
    src_file = os.path.join(subjects_dir, subject, 'bem',
                            '%s-%s-src.fif' % (subject, src_tag))
    src, spacing = src_tag.split('-')
    if os.path.exists(src_file):
        return mne.read_source_spaces(src_file, False)
    elif src == 'ico':
        ss = mne.setup_source_space(subject, spacing=src + spacing,
                                    subjects_dir=subjects_dir, add_dist=True)
    elif src == 'vol':
        mri_file = os.path.join(subjects_dir, subject, 'mri', 'orig.mgz')
        bem_file = os.path.join(subjects_dir, subject, 'bem',
                                'sample-5120-5120-5120-bem-sol.fif')
        ss = mne.setup_volume_source_space(subject, pos=float(spacing),
                                           mri=mri_file, bem=bem_file,
                                           mindist=0., exclude=0.,
                                           subjects_dir=subjects_dir)
    else:
        raise ValueError("src_tag=%s" % repr(src_tag))
    mne.write_source_spaces(src_file, ss)
    return ss


def get_mne_sample(tmin=-0.1, tmax=0.4, baseline=(None, 0), sns=False,
                   src=None, sub="modality=='A'", ori='free', snr=2,
                   method='dSPM', rm=False, stc=False, hpf=0):
    """Load events and epochs from the MNE sample data

    Parameters
    ----------
    tmin : scalar
        Relative time of the first sample of the epoch.
    tmax : scalar
        Relative time of the last sample of the epoch.
    baseline : {None, tuple of 2 {scalar, None}}
        Period for baseline correction.
    sns : bool | str
        Add sensor space data as NDVar as ``ds['meg']`` (default ``False``).
        Set to ``'grad'`` to load gradiometer data.
    src : False | 'ico' | 'vol'
        Add source space data as NDVar as ``ds['src']`` (default ``False``).
    sub : str | list | None
        Expression for subset of events to load. For a very small dataset use e.g.
        ``[0,1]``.
    ori : 'free' | 'fixed' | 'vector'
        Orientation of sources.
    snr : scalar
        MNE inverse parameter.
    method : str
        MNE inverse parameter.
    rm : bool
        Pretend to be a repeated measures dataset (adds 'subject' variable).
    stc : bool
        Add mne SourceEstimate for source space data as ``ds['stc']`` (default
        ``False``).
    hpf : scalar
        High pass filter cutoff.

    Returns
    -------
    ds : Dataset
        Dataset with epochs from the MNE sample dataset in ``ds['epochs']``.
    """
    if ori == 'free':
        loose = 1
        fixed = False
        pick_ori = None
    elif ori == 'fixed':
        loose = 0
        fixed = True
        pick_ori = None
    elif ori == 'vector':
        if LooseVersion(mne.__version__) < LooseVersion('0.17'):
            raise RuntimeError(f'mne version {mne.__version__}; vector source estimates require mne 0.17')
        loose = 1
        fixed = False
        pick_ori = 'vector'
    else:
        raise ValueError(f"ori={ori!r}")

    data_dir = mne.datasets.sample.data_path()
    meg_dir = os.path.join(data_dir, 'MEG', 'sample')
    raw_file = os.path.join(meg_dir, 'sample_audvis_filt-0-40_raw.fif')
    event_file = os.path.join(meg_dir, 'sample_audvis_filt-0-40-eve.fif')
    subjects_dir = os.path.join(data_dir, 'subjects')
    subject = 'sample'
    label_path = os.path.join(subjects_dir, subject, 'label', '%s.label')

    if not os.path.exists(event_file):
        raw = mne.io.Raw(raw_file)
        events = mne.find_events(raw, stim_channel='STI 014')
        mne.write_events(event_file, events)
    ds = load.fiff.events(raw_file, events=event_file)
    if hpf:
        ds.info['raw'].load_data()
        ds.info['raw'].filter(hpf, None)
    ds.index()
    ds.info['subjects_dir'] = subjects_dir
    ds.info['subject'] = subject
    ds.info['label'] = label_path

    # get the trigger variable form the dataset for eaier access
    trigger = ds['trigger']

    # use trigger to add various labels to the dataset
    ds['condition'] = Factor(trigger, labels={
        1: 'LA', 2: 'RA', 3: 'LV', 4: 'RV', 5: 'smiley', 32: 'button'})
    ds['side'] = Factor(trigger, labels={
        1: 'L', 2: 'R', 3: 'L', 4: 'R', 5: 'None', 32: 'None'})
    ds['modality'] = Factor(trigger, labels={
        1: 'A', 2: 'A', 3: 'V', 4: 'V', 5: 'None', 32: 'None'})

    if rm:
        ds = ds.sub('trigger < 5')
        ds = ds.equalize_counts('side % modality')
        subject_f = ds.eval('side % modality').enumerate_cells()
        ds['subject'] = subject_f.as_factor('s%r', random=True)

    if sub:
        ds = ds.sub(sub)

    load.fiff.add_mne_epochs(ds, tmin, tmax, baseline)
    if sns:
        ds['meg'] = load.fiff.epochs_ndvar(ds['epochs'],
                                           data='mag' if sns is True else sns,
                                           sysname='neuromag306')

    if not src:
        return ds
    elif src == 'ico':
        src_tag = 'ico-4'
    elif src == 'vol':
        src_tag = 'vol-10'
    else:
        raise ValueError(f"src={src!r}")
    epochs = ds['epochs']

    # get inverse operator
    inv_file = os.path.join(meg_dir, f'sample_eelbrain_{src_tag}-inv.fif')
    if os.path.exists(inv_file):
        inv = mne.minimum_norm.read_inverse_operator(inv_file)
    else:
        fwd_file = os.path.join(meg_dir, 'sample-%s-fwd.fif' % src_tag)
        bem_dir = os.path.join(subjects_dir, subject, 'bem')
        bem_file = os.path.join(bem_dir, 'sample-5120-5120-5120-bem-sol.fif')
        trans_file = os.path.join(meg_dir, 'sample_audvis_raw-trans.fif')

        if os.path.exists(fwd_file):
            fwd = mne.read_forward_solution(fwd_file)
        else:
            src_ = _mne_source_space(subject, src_tag, subjects_dir)
            fwd = mne.make_forward_solution(epochs.info, trans_file, src_, bem_file)
            mne.write_forward_solution(fwd_file, fwd)

        cov_file = os.path.join(meg_dir, 'sample_audvis-cov.fif')
        cov = mne.read_cov(cov_file)
        inv = mn.make_inverse_operator(epochs.info, fwd, cov, loose=loose,
                                       depth=None, fixed=fixed)
        mne.minimum_norm.write_inverse_operator(inv_file, inv)
    ds.info['inv'] = inv

    stcs = mn.apply_inverse_epochs(epochs, inv, 1. / (snr ** 2), method,
                                   pick_ori=pick_ori)
    ds['src'] = load.fiff.stc_ndvar(stcs, subject, src_tag, subjects_dir,
                                    method, fixed)
    if stc:
        ds['stc'] = stcs

    return ds


def get_ndvar(case=0, time=100, frequency=8, cat=0, sensor=0, name='ndvar'):
    dims = []
    if case:
        dims.append(Case(case))
    if time:
        dims.append(UTS(-0.1, 0.01, time))
    if frequency:
        dims.append(Scalar('frequency', np.logspace(2, 3.5, frequency)))
    if cat:
        dims.append(Categorial('cat', string.ascii_lowercase[:cat]))
    if sensor:
        dims.append(get_sensor(sensor))
    shape = [len(dim) for dim in dims]
    x = np.random.normal(0, 1, shape)
    return NDVar(x, dims, name)


def get_uts(utsnd=False, seed=0, nrm=False, vector3d=False):
    """Create a sample Dataset with 60 cases and random data.

    Parameters
    ----------
    utsnd : bool
        Add a sensor by time NDVar (called 'utsnd').
    seed : None | int
        If not None, call ``numpy.random.seed(seed)`` to ensure replicability.
    nrm : bool
        Add a nested random effect Factor "nrm" (nested in ``A``).
    vector3d : bool
        Add a space x time vector time series as ``v3d``.

    Returns
    -------
    ds : Dataset
        Datasets with data from random distributions.
    """
    random = np.random if seed is None else np.random.RandomState(seed)

    ds = Dataset()

    # add a model
    ds['A'] = Factor(['a0', 'a1'], repeat=30)
    ds['B'] = Factor(['b0', 'b1'], repeat=15, tile=2)
    ds['rm'] = Factor(('R%.2i' % i for i in range(15)), tile=4, random=True)
    ds['ind'] = Factor(('R%.2i' % i for i in range(60)), random=True)

    # add dependent variables
    rm_var = np.tile(random.normal(size=15), 4)
    y = np.hstack((random.normal(size=45), random.normal(1, size=15)))
    y += rm_var
    ds['Y'] = Var(y)
    ybin = random.randint(0, 2, size=60)
    ds['YBin'] = Factor(ybin, labels={0: 'c1', 1: 'c2'})
    ycat = random.randint(0, 3, size=60)
    ds['YCat'] = Factor(ycat, labels={0: 'c1', 1: 'c2', 2: 'c3'})

    # add a uts NDVar
    time = UTS(-.2, .01, 100)
    y = random.normal(0, .5, (60, len(time)))
    y += rm_var[:, None]
    y[:15, 20:60] += np.hanning(40) * 1  # interaction
    y[:30, 50:80] += np.hanning(30) * 1  # main effect
    ds['uts'] = NDVar(y, ('case', time))

    # add sensor NDVar
    if utsnd:
        y = random.normal(0, 1, (60, 5, len(time)))
        y += rm_var[:, None, None]
        # add interaction
        win = np.hanning(50)
        y[:15, 0, 50:] += win * 3
        y[:15, 1, 50:] += win * 2
        y[:15, 4, 50:] += win
        # add main effect
        y[30:, 2, 25:75] += win * 2.5
        y[30:, 3, 25:75] += win * 1.5
        y[30:, 4, 25:75] += win
        # add spectral effect
        freq = 15.0  # >= 2
        x = np.sin(time.times * freq * 2 * np.pi)
        for i in range(30):
            shift = random.randint(0, 100 / freq)
            y[i, 2, 25:75] += 1.1 * win * x[shift: 50 + shift]
            y[i, 3, 25:75] += 1.5 * win * x[shift: 50 + shift]
            y[i, 4, 25:75] += 0.5 * win * x[shift: 50 + shift]

        sensor = get_sensor(5)
        dims = ('case', sensor, time)
        ds['utsnd'] = NDVar(y, dims, info=_info.for_eeg())

    # nested random effect
    if nrm:
        ds['nrm'] = Factor([f'{a}{i:02}' for a in 'AB' for _ in range(2) for
                            i in range(15)], random=True)

    if vector3d:
        x = random.normal(0, 1, (60, 3, 100))
        # main effect
        x[:30, 0, 50:80] += np.hanning(30) * 0.7
        x[:30, 1, 50:80] += np.hanning(30) * -0.5
        x[:30, 2, 50:80] += np.hanning(30) * 0.3
        ds['v3d'] = NDVar(x, (Case, Space('RAS'), time))

    return ds


def get_uv(seed=0, nrm=False, vector=False):
    """Dataset with random univariate data

    Parameters
    ----------
    seed : None | int
        Seed the numpy random state before generating random data.
    nrm : bool
        Add a nested random-effects variable (default False).
    vector : bool
        Add a 3d vector variable as ``ds['v']`` (default ``False``).
    """
    random = np.random if seed is None else np.random.RandomState(seed)

    ds = permute([('A', ('a1', 'a2')),
                  ('B', ('b1', 'b2')),
                  ('rm', ['s%03i' % i for i in range(20)])])
    ds['rm'].random = True
    ds['intvar'] = Var(random.randint(5, 15, 80))
    ds['intvar'][:20] += 3
    ds['fltvar'] = Var(random.normal(0, 1, 80))
    ds['fltvar'][:40] += 1.
    ds['fltvar2'] = Var(random.normal(0, 1, 80))
    ds['fltvar2'][40:] += ds['fltvar'][40:].x
    ds['index'] = Var(np.repeat([True, False], 40))
    if nrm:
        ds['nrm'] = Factor(['s%03i' % i for i in range(40)], tile=2, random=True)
    if vector:
        x = random.normal(0, 1, (80, 3))
        x[:40] += [.3, .3, .3]
        ds['v'] = NDVar(x, (Case, Space('RAS')))
    return ds


def get_sensor(n):
    assert n == 5
    locs = np.array([[-1.0, 0.0, 0.0],
                     [0.0, 1.0, 0.0],
                     [1.0, 0.0, 0.0],
                     [0.0, -1.0, 0.0],
                     [0.0, 0.0, 1.0]])
    sensor = Sensor(locs, sysname='test_sens')
    sensor.set_connectivity(connect_dist=1.75)
    return sensor


def setup_samples_experiment(
        dst: PathArg,
        n_subjects: int = 3,
        n_segments: int = 4,
        n_sessions: int = 1,
        n_visits: int = 1,
        name: str = 'SampleExperiment',
        mris: bool = False,
        mris_only: bool = False,
        pick: str = 'mag',
):
    """Setup up file structure for the ``SampleExperiment`` class

    Corresponding to the pipeline in ``examples/experiment``.

    Parameters
    ----------
    dst
        Path. ``dst`` should exist, a new folder called ``name`` will be
        created within ``dst``.
    n_subjects
        Number of subjects.
    n_segments
        Number of data segments to include in each file.
    n_sessions
        Number of sessions.
    n_visits
        Number of visits.
    name
        Name for the directory for the new experiment (default
        ``'SampleExperiment'``).
    mris
        Set up MRIs.
    mris_only
        Only create MRIs, skip MEG data (add MRIs to existing experiment data).
    pick
        Pick a certain channel type (``''`` to copy all channels).
    """
    # find data source
    data_path = Path(mne.datasets.sample.data_path())
    fsaverage_path = Path(mne.datasets.fetch_fsaverage())

    # setup destination
    dst = Path(dst).expanduser().resolve()
    root = dst / name
    root.mkdir(exist_ok=mris_only)

    if n_sessions > 1 and n_visits > 1:
        raise NotImplementedError
    n_recordings = n_subjects * max(n_sessions, n_visits)
    subjects = [f'R{s_id:04}' for s_id in range(n_subjects)]

    meg_sdir = root / 'meg'
    meg_sdir.mkdir(exist_ok=mris_only)

    if mris:
        mri_sdir = root / 'mri'
        if mris_only and mri_sdir.exists():
            shutil.rmtree(mri_sdir)
        mri_sdir.mkdir()
        # copy rudimentary fsaverage
        surf_names = ['inflated', 'white', 'orig', 'orig_avg', 'curv', 'sphere']
        files = {
            'bem': ['fsaverage-head.fif', 'fsaverage-inner_skull-bem.fif'],
            'label': ['lh.aparc.annot', 'rh.aparc.annot'],
            'surf': [f'{hemi}.{name}' for hemi, name in product(['lh', 'rh'], surf_names)],
            'mri': [],
        }
        dst_s_dir = mri_sdir / 'fsaverage'
        dst_s_dir.mkdir()
        # from fsaverage
        for dir_name, file_names in files.items():
            src_dir = fsaverage_path / dir_name
            dst_dir = dst_s_dir / dir_name
            dst_dir.mkdir()
            for file_name in file_names:
                shutil.copy(src_dir / file_name, dst_dir / file_name)
        # source space
        src_src = fsaverage_path / 'bem' / 'fsaverage-ico-3-src.fif'
        src_dst = dst_s_dir / 'bem' / 'fsaverage-ico-3-src.fif'
        if not src_src.exists():
            src = mne.setup_source_space('fsaverage', 'ico1', subjects_dir=fsaverage_path.parent)
            src.save(src_src)
        shutil.copy(src_src, src_dst)
        # create scaled brains
        trans = mne.Transform(4, 5, [[ 0.9998371,  -0.00766024,  0.01634169,  0.00289569],
                                     [ 0.00933457,  0.99443108, -0.10497498, -0.0205526 ],
                                     [-0.01544655,  0.10511042,  0.9943406,  -0.04443745],
                                     [ 0.,          0.,          0.,          1.        ]])
        # os.environ['_MNE_FEW_SURFACES'] = 'true'
        for subject in subjects:
            mne.scale_mri('fsaverage', subject, 1., subjects_dir=mri_sdir, skip_fiducials=True, labels=False)
            meg_dir = meg_sdir / subject
            meg_dir.mkdir(exist_ok=mris_only)
            trans.save(str(meg_dir / f'{subject}-trans.fif'))
        # del os.environ['_MNE_FEW_SURFACES']
    if mris_only:
        return

    # MEG
    raw_path = data_path / 'MEG' / 'sample' / 'sample_audvis_raw.fif'
    raw = mne.io.read_raw_fif(str(raw_path))
    raw.info['bads'] = []
    sfreq = raw.info['sfreq']

    # find segmentation points
    events = mne.find_events(raw)
    events[:, 0] -= raw.first_samp
    segs = []
    n = 0
    t_start = 0
    for sample, _, trigger in events:
        if trigger == 5:  # smiley
            n += 1
        if n == n_segments:
            t = sample / sfreq
            segs.append((t_start, t))
            if len(segs) == n_recordings:
                break
            t_start = t
            n = 0
    else:
        raise ValueError("Not enough data in sample raw. Try smaller ns.")

    if n_visits > 1:
        sessions = ['sample', *(f'sample {i}' for i in range(1, n_visits))]
    elif n_sessions > 1:
        sessions = ['sample%i' % (i + 1) for i in range(n_sessions)]
    else:
        sessions = ['sample']

    for subject in subjects:
        meg_dir = meg_sdir / subject
        meg_dir.mkdir(exist_ok=mris)
        for session in sessions:
            start, stop = segs.pop()
            raw_ = raw.copy().crop(start, stop)
            raw_.load_data()
            if pick:
                raw_.pick_types(pick, stim=True, exclude=[])
            raw_.save(str(meg_dir / f'{subject}_{session}-raw.fif'))



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
from os.path import join
from warnings import catch_warnings, filterwarnings

import pytest

from eelbrain import *
from eelbrain.pipeline import *
from eelbrain.testing import TempDir, path, requires_mne_sample_data


class Experiment(MneExperiment):

    meg_system = 'neuromag306mag'

    sessions = 'sample'

    visits = ('', '1')

    raw = {
        '1-40': RawFilter('raw', 1, 40, method='iir'),
        'ica': RawICA('raw', 'sample', 'fastica', max_iter=1),
        'ica1-40': RawFilter('ica', 1, 40, method='iir'),
    }

    variables = {
        'event': {(1, 2, 3, 4): 'target', 5: 'smiley', 32: 'button'},
        'side': {(1, 3): 'left', (2, 4): 'right'},
        'modality': {(1, 2): 'auditory', (3, 4): 'visual'}
    }

    epochs = {
        'target': PrimaryEpoch('sample', "event == 'target'", tmax=0.15, decim=4),
    }

    tests = {
        'side': TTestRelated('side', 'left', 'right'),
    }


def test_visit_patterns():
    e = Experiment()
    assert e._glob_pattern('fwd-file', True, session='sample') == path('/eelbrain-cache/raw/*/sample*-*-*-fwd.fif')


@requires_mne_sample_data
def test_mne_experiment_visit():
    set_log_level('warning', 'mne')
    tempdir = TempDir()
    datasets.setup_samples_experiment(tempdir, 3, 1, n_visits=2)
    root = join(tempdir, 'SampleExperiment')

    e = Experiment(root)
    ds0 = e.load_events(epoch='target', visit='')
    ds1 = e.load_events(epoch='target', visit='1')
    assert ds0[1, 'trigger'] != ds1[1, 'trigger']
    for _ in e:
        e.make_epoch_selection(auto=2e-12)
    e.set(visit='')
    for _ in e:
        e.make_epoch_selection(auto=2e-12)
    res0 = e.load_test('side', 0.05, 0.15, 0.05, data='sensor', make=True)
    e.set(visit='1')
    with pytest.raises(IOError):
        e.load_test('side', 0.05, 0.15, 0.05, data='sensor')
    res1 = e.load_test('side', 0.05, 0.15, 0.05, data='sensor', make=True)
    assert res1.p.min() != res0.p.min()

    # ica
    e.set(raw='ica')
    with catch_warnings():
        filterwarnings('ignore', "FastICA did not converge", UserWarning)
        assert e.make_ica(visit='') == join(root, 'meg', 'R0000', 'R0000 ica-ica.fif')
        assert e.make_ica(visit='1') == join(root, 'meg', 'R0000', 'R0000 1 ica-ica.fif')



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
from .. import plot
from .._data_obj import Dataset
from .._stats.testnd import _MergedTemporalClusterDist, NDTest
from .._text import ms
from ..fmtxt import Section


def time_results(
        res: NDTest,
        ds: Dataset,
        colors: dict,
        title: str = 'Results',
        caption: str = "Timecourse",
        pairwise_pmax: float = 0.1,
        merged_dist: _MergedTemporalClusterDist = None,
):
    """Add time course with clusters

    Parameters
    ----------
    res
        Result of the temporal cluster test.
    ds
        Data.
    colors
        Cell colors.
    title
        Section title.
    caption
        Time-course figure caption.
    pairwise_pmax
        Barplots for clusters with ``p <= pairwise_pmax``.
    merged_dist
        Merged cluster distribution for correcting p values across ROIs
    """
    if merged_dist:
        clusters = merged_dist.correct_cluster_p(res)
    else:
        clusters = res.find_clusters()
    if clusters.n_cases:
        idx = clusters.eval("p.argmin()")
        max_sig = clusters['sig'][idx]
        if max_sig:
            title += max_sig
    section = Section(title)

    # compose captions
    if clusters.n_cases:
        c_caption = f"Clusters in time window {res._desc_timewindow()} based on {res._desc_samples}."
        if 'p_parc' in clusters:
            c_caption += " p: p-value in ROI; p_parc: p-value corrected across ROIs."
        tc_caption = caption
    else:
        c_caption = f"No clusters found {res._desc_timewindow()}."
        tc_caption = ' '.join((caption, c_caption))

    # plotting arguments
    model = res._plot_model()
    sub = res._plot_sub()

    # add UTSStat plot
    p = plot.UTSStat(res.y, model, None, res.match, sub, ds, colors=colors, legend=False, clusters=clusters, show=False)
    ax = p.axes[0]
    if res.tstart is not None:
        ax.axvline(res.tstart, color='k')
    if res.tstop is not None:
        ax.axvline(res.tstop, color='k')
    image = p.image('%s_cluster.png')
    legend_p = p.plot_legend(show=False)
    legend = legend_p.image("Legend")
    section.add_figure(tc_caption, [image, legend])
    p.close()
    legend_p.close()

    # add cluster table
    if clusters.n_cases:
        t = clusters.as_table(midrule=True, caption=c_caption)
        section.append(t)

    # pairwise plots
    model_ = model
    colors_ = colors
    if pairwise_pmax is not None:
        plots = []
        clusters_ = clusters.sub("p <= %s" % pairwise_pmax)
        clusters_.sort("tstart")
        for cluster in clusters_.itercases():
            cid = cluster['id']
            c_tstart = cluster['tstart']
            c_tstop = cluster['tstop']
            tw_str = "%s - %s ms" % (ms(c_tstart), ms(c_tstop))
            if 'effect' in cluster:
                title = "%s %s%s: %s" % (cluster['effect'], cid, cluster['sig'], tw_str)
                model_ = cluster['effect'].replace(' x ', '%')
                colors_ = colors if model_ == model else None
            else:
                title = "Cluster %s%s: %s" % (cid, cluster['sig'], tw_str)
            y_ = ds[res.y].summary(time=(c_tstart, c_tstop))
            p = plot.Barplot(y_, model_, res.match, sub, ds=ds, corr=None, show=False, colors=colors_, title=title)
            plots.append(p.image())
            p.close()

        section.add_image_figure(plots, "Value in the time-window of the clusters with uncorrected pairwise t-tests.")

    return section



"""Plotting for data-objects"""
# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>

from .._colorspaces import two_step_colormap

from ._base import reset_rc
from ._colors import ColorBar, ColorGrid, ColorList
from ._figure import Figure
from ._glassbrain import GlassBrain
from ._line import LineStack
from ._sensors import SensorMaps, SensorMap
from ._styles import Style, colors_for_categorial, colors_for_oneway, colors_for_twoway, single_hue_colormap, soft_threshold_colormap
from ._topo import TopoArray, TopoButterfly, Topomap, TopomapBins
from ._split import DataSplit
from ._uts import UTSStat, UTS, UTSClusters
from ._utsnd import Array, Butterfly
from ._uv import Barplot, BarplotHorizontal, Boxplot, Scatter, Histogram, PairwiseLegend, Regression, Timeplot
from . import brain



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""Plot sensor maps."""
from math import sin, cos, asin
import os
from typing import Literal, Sequence, Union

import numpy as np
import matplotlib as mpl
import matplotlib.axes
from matplotlib.lines import Line2D

from .._data_obj import Datalist, Sensor, as_sensor
from ._base import EelFigure, ImLayout, ColorArg

SENSORMAP_FRAME = 0.1

# some useful kwarg dictionaries for different plot layouts
kwargs_mono = dict(mc='k',
                   lc='.5',
                   hllc='k',
                   hlmc='k',
                   hlms=7,
                   strlc='k')


def _head_outlines(
        radius: Union[float, Sequence[float]],
        center: Union[float, Sequence[float]] = 0,
):
    # generate outlines for center 0, radius 1
    nose_alpha = 0.2
    l = np.linspace(0, 2 * np.pi, 101)
    head_x = np.cos(l)
    head_y = np.sin(l)
    w = sin(nose_alpha)
    nose_x = np.array((-w, -w * 0.5, -w * 0.2, 0, w * 0.2, w * 0.5, w))
    ymin = cos(nose_alpha)
    nose_y = np.array((ymin, 1.02, 1.09, 1.1, 1.09, 1.02, ymin))
    ear_y = np.array((0.15, 0.145, 0.135, 0.125, 0.111, -0.011, -0.1864,
                      -0.2626, -0.2768, -0.2398))
    ear_x_right = np.array((cos(asin(ear_y[0])), 1., 1.02, 1.025, 1.03, 1.04,
                            1.07, 1.06, 1.02, cos(asin(ear_y[-1]))))
    ear_x_left = -ear_x_right

    # apply radius and center
    if isinstance(radius, Sequence):
        rx, ry = radius
    else:
        rx = ry = radius

    if isinstance(center, Sequence):
        cx, cy = center
        cx += 0.5
        cy += 0.5
    else:
        cx = 0.5
        cy = center + 0.5

    for item in (head_x, nose_x, ear_x_right, ear_x_left):
        item *= rx
        item += cx

    for item in (head_y, nose_y, ear_y):
        item *= ry
        item += cy

    return ((head_x, head_y), (nose_x, nose_y), (ear_x_left, ear_y),
            (ear_x_right, ear_y))


class _plt_connectivity:
    def __init__(self, ax, locs, connectivity, linestyle={}):
        self.ax = ax
        self.locs = locs
        self._h = []
        self.show(connectivity, linestyle)

    def show(self, connectivity, linestyle={}):
        while self._h:
            self._h.pop().remove()

        if connectivity is None:
            return

        for c, r in connectivity:
            x = self.locs[[c, r], 0]
            y = self.locs[[c, r], 1]
            line = Line2D(x, y, **linestyle)
            self.ax.add_line(line)
            self._h.append(line)


class _ax_map2d:

    def __init__(self, ax, sensors, proj, extent, size, color, marker,
                 mark=None, head_radius=None, head_pos=0., head_linewidth=None,
                 labels='none'):
        self.ax = ax

        self.sensors = _plt_map2d(ax, sensors, proj, extent, marker, size,
                                  color, mark, None, None, labels, True,
                                  head_radius, head_pos, head_linewidth)

        locs = sensors.get_locs_2d(proj, extent, SENSORMAP_FRAME)
        self.connectivity = _plt_connectivity(ax, locs, None)

        ax.set_aspect('equal', 'datalim', 'C')
        if extent:
            ax.set_xlim(0, extent)
            ax.set_ylim(0, extent)

    def mark_sensors(self, *args, **kwargs):
        self.sensors.mark_sensors(*args, **kwargs)

    def remove(self):
        "Remove from axes"
        self.sensors.remove()


class _plt_map2d:
    """Sensor-map plot

    Parameters
    ----------
    ax
        Axes.
    sensors
        Sensor dimension.
    """
    _label_text = 'none'  # currently shown label text

    def __init__(
            self,
            ax: matplotlib.axes.Axes,
            sensors: Sensor,
            proj: str,
            extent: float,
            marker: str,
            size: float,
            color: ColorArg,
            mark: Union[Sequence[str], str, Sequence[int], int],
            mcolor: ColorArg,
            mmarker: str,
            labels: Literal['', 'none', 'index', 'name', 'fullname'],
            invisible: bool,
            head_radius: Union[float, Sequence[float]],
            head_pos: Union[float, Sequence[float]],
            head_linewidth: float = None,
    ):
        self.ax = ax
        self.sensors = sensors
        self.locs = sensors.get_locs_2d(proj, extent, SENSORMAP_FRAME)
        self._index = None if invisible else sensors._visible_sensors(proj)
        self._extent = extent
        self._label_h = []  # handles for label text objects
        self._mark_handles = []

        # head outline
        if head_radius:
            if head_radius is True:
                head_radius = 0.5 * (1 - (SENSORMAP_FRAME * 0.9))

            if head_linewidth is None:
                head_linewidth = mpl.rcParams['lines.linewidth']

            for x, y in _head_outlines(head_radius, head_pos):
                ax.plot(x, y, color='k', linewidth=head_linewidth, solid_capstyle='butt', clip_on=False)

        # sensors
        if labels == '':
            return
        index = slice(None) if self._index is None else self._index
        self._sensor_h = ax.scatter(self.locs[index, 0], self.locs[index, 1], size, color, marker)

        if labels:
            self.show_labels(labels)

        if mark is not None:
            self.mark_sensors(mark, 20, mcolor, mmarker)

    def mark_sensors(self, sensors, s=20, c='yellow', marker='o',
                     *args, **kwargs):
        """Mark specific sensors

        Parameters
        ----------
        sensors : None | Sensor dimension index
            Sensors which should be marked (None to clear all markings).
        s : scalar | sequence of scalars
            Marker size(s) in points^2 (default 20).
        c : color | sequence of colors
            Marker color(s) (default ``'yellow'``).
        marker : str
            Marker style (default: ``'o'``).
        ... :
            Matplotlib :func:`~matplotlib.axes.Axes.scatter` parameters.
        """
        if sensors is None:
            while self._mark_handles:
                self._mark_handles.pop().remove()
            return

        if c is None:
            c = 'yellow'
        if marker is None:
            marker = 'o'

        idx = self.sensors._array_index(sensors)
        h = self.ax.scatter(self.locs[idx, 0], self.locs[idx, 1], s, c, marker,
                            *args, **kwargs)
        self._mark_handles.append(h)

    def remove(self):
        "Remove from axes"
        self._sensor_h.remove()
        while self._mark_handles:
            self._mark_handles.pop().remove()

    def show_labels(self, text='name', xpos=0, ypos=0, ha='center', va='bottom',
                    **text_kwargs):
        """Plot labels for the sensors

        Parameters
        ----------
        labels : 'none' | 'index' | 'name' | 'fullname'
            Content of the labels. For 'name', any prefix common to all names
            is removed; with 'fullname', the full name is shown.
        xpos, ypos : scalar
            The position offset of the labels from the sensor markers.
        text_kwargs : **
            Matplotlib text parameters.
        """
        if not text:
            text = 'none'
        elif text not in ('none', 'index', 'name', 'fullname'):
            raise ValueError("Invalid sensor label argument: %s. Need one of "
                             "'none' | 'index' | 'name' | 'fullname'" %
                             repr(text))

        if text == self._label_text:
            return
        self._label_text = text

        # remove existing labels
        while self._label_h:
            self._label_h.pop().remove()

        if text == 'none':
            return
        elif text == 'index':
            labels = list(map(str, range(len(self.sensors))))
        elif text == 'name':
            labels = self.sensors.names
            prefix = os.path.commonprefix(labels)
            pf_len = len(prefix)
            labels = [label[pf_len:] for label in labels]
        elif text == 'fullname':
            labels = self.sensors.names
        else:
            raise RuntimeError("text=%s" % repr(text))

        locs = self.locs
        if self._index is not None:
            labels = Datalist(labels)[self._index]
            locs = locs[self._index]

        locs = locs + [[xpos, ypos]]
        for (x, y), txt in zip(locs, labels):
            h = self.ax.text(x, y, txt, ha=ha, va=va, **text_kwargs)
            self._label_h.append(h)

    def separate_labels(self, pad):
        # separate overlapping labels
        locs = self.locs
        data_to_screen = self.ax.transData.transform_point
        screen_to_data = self.ax.transData.inverted().transform_point

        if self._extent:
            center = self._extent / 2.
            label_order = np.argsort((locs[:, 0] - center) ** 2)
        else:
            center = 0
            label_order = np.argsort(locs[:, 0] ** 2)

        n_labels = len(label_order)
        for i in range(1, n_labels):
            i_i = label_order[i]
            i_loc = locs[i_i, 0]
            if i_loc == center:
                continue
            h = self._label_h[i_i]
            i_bbox = h.get_window_extent()
            padded_bbox = i_bbox.expanded(pad, 0)
            side = -1 if i_loc < center else 1

            dx = 0
            for j in range(i):
                j_i = label_order[j]
                j_bbox = self._label_h[j_i].get_window_extent()
                if padded_bbox.overlaps(j_bbox):
                    if side == -1:
                        dx = min(dx, j_bbox.xmin - i_bbox.xmax - pad)
                    else:
                        dx = max(dx, j_bbox.xmax - i_bbox.xmin + pad)

            if dx:
                x, y = data_to_screen(h.get_position())
                h.set_position(screen_to_data((x + dx, y)))

    def set_label_color(self, color='k'):
        """Change the color of all sensor labels

        Parameters
        ----------
        color : matplotlib color
            New color for the sensor labels.
        """
        for h in self._label_h:
            h.set_color(color)


class SensorMapMixin:
    # expects self._sensor_plots to be list of _plt_map2d
    __label_options = ['None', 'Index', 'Name', 'Full Name']
    __label_option_args = ['none', 'index', 'name', 'fullname']

    def __init__(self, sensor_plots):
        """Call after EelFigure init (toolbar fill)

        Parameters
        ----------
        sensor_plots : list of _plt_map2d
            Sensor-map objects.
        """
        self.__label_color = 'k'
        self.__sensor_plots = sensor_plots
        if not self._has_frame:
            return
        # find current label text
        sel = self.__label_option_args.index(sensor_plots[0]._label_text)
        self.__LabelChoice.SetSelection(sel)

    def _fill_toolbar(self, tb):
        from .._wxgui import wx

        # sensor labels
        lbl = wx.StaticText(tb, -1, "Labels:")
        tb.AddControl(lbl)
        choice = wx.Choice(tb, -1, choices=self.__label_options)
        tb.AddControl(choice)
        self.__LabelChoice = choice
        choice.Bind(wx.EVT_CHOICE, self.__OnSensorLabelChoice)

        # sensor label color
        choices = ['black', 'white', 'blue', 'green', 'red', 'cyan', 'magenta',
                   'yellow']
        choice = wx.Choice(tb, -1, choices=choices)
        tb.AddControl(choice)
        self.__LabelColorChoice = choice
        choice.Bind(wx.EVT_CHOICE, self.__OnSensorLabelColorChoice)

        btn = wx.Button(tb, label="Mark")  # , style=wx.BU_EXACTFIT)
        btn.Bind(wx.EVT_BUTTON, self.__OnMarkSensor)
        tb.AddControl(btn)

    def __OnMarkSensor(self, event):
        from .._wxgui import wx

        msg = "Channels to mark, separated by comma"
        dlg = wx.TextEntryDialog(self._frame, msg, "Mark Sensor")
        if dlg.ShowModal() != wx.ID_OK:
            return

        chs = tuple(filter(None, map(str.strip, dlg.GetValue().split(','))))
        try:
            self.mark_sensors(chs)
        except Exception as exc:
            msg = '%s: %s' % (type(exc).__name__, exc)
            sty = wx.OK | wx.ICON_ERROR
            wx.MessageBox(msg, "Mark Sensors Failed for %r" % chs, style=sty)

    def __OnSensorLabelChoice(self, event):
        sel = event.GetSelection()
        sel_arg = self.__label_option_args[sel]
        self.set_label_text(sel_arg)

    def __OnSensorLabelColorChoice(self, event):
        sel = event.GetSelection()
        color = ['k', 'w', 'b', 'g', 'r', 'c', 'm', 'y'][sel]
        self.set_label_color(color)

    def mark_sensors(self, sensors, axis=None, s=20, c='yellow', marker='o', *args, **kwargs):
        """Mark given sensors on the plots

        Parameters
        ----------
        sensors : None | Sensor dimension index
            Sensors which should be marked (None to clear all markings).
        axis : int | list of int
            Which axes to mark (default is all).
        s : scalar | sequence of scalars
            Marker size(s) in points^2 (default 20).
        c : color | sequence of colors
            Marker color(s) (default ``'yellow'``).
        marker : str
            Marker style (default: ``'o'``).
        ... :
            Matplotlib :func:`~matplotlib.axes.Axes.scatter` parameters.
        """
        if axis is None:
            plots = self.__sensor_plots
        elif isinstance(axis, int):
            plots = [self.__sensor_plots[axis]]
        else:
            plots = [self.__sensor_plots[i] for i in axis]

        for p in plots:
            p.mark_sensors(sensors, s, c, marker, *args, **kwargs)
        self.draw()

    def separate_labels(self, pad=10):
        """Move overlapping labels apart along the x axis

        Parameters
        ----------
        pad : scalar
            Minimum amount of padding between labels (in pixels; default 5).
        """
        for p in self.__sensor_plots:
            p.separate_labels(pad)
        self.draw()

    def set_label_color(self, color='w'):
        if hasattr(self, '_SensorLabelChoice'):
            sels = ['k', 'w', 'b', 'g', 'r', 'c', 'm', 'y']
            if color in sels:
                sel = sels.index(color)
                self.__LabelColorChoice.SetSelection(sel)

        self.__label_color = color
        for p in self.__sensor_plots:
            p.set_label_color(color)
        self.draw()

    def set_label_text(self, text='name'):
        """Add/remove sensor labels

        Parameters
        ----------
        labels : 'none' | 'index' | 'name' | 'fullname'
            Content of the labels. For 'name', any prefix common to all names
            is removed; with 'fullname', the full name is shown.
        """
        # update plots
        for p in self.__sensor_plots:
            p.show_labels(text, color=self.__label_color)
        # update GUI
        self.__LabelChoice.SetSelection(self.__label_option_args.index(text))
        self.draw()


class SensorMaps(EelFigure):
    """Multiple views on a sensor layout.

    Allows selecting sensor groups and retrieving corresponding indices.

    Parameters
    ----------
    sensors : Sensor | NDVar
        The :class:`Sensor` dimension, or an :class:`NDVar` with a sensor
        dimension.
    select : list of int
        Initial selection.
    proj : str
        Sensor projection for the fourth plot.
    size : scalar
        Size for the sensor markers.
    color : matplotlib color
        Color for the sensor markers.
    marker : str
        Marker for the sensor positions.
    frame : scalar
        Size of the empty space around sensors in axes.
    ...
        Also accepts :ref:`general-layout-parameters`.

    Notes
    -----
    **Selecting Sensor Groups:**

     - Dragging with the left mouse button adds sensors to the selection.
     - Dragging with the right mouse button removes sensors from the current
       selection.
     - The 'Clear' button (or :meth:`clear`) clears the selection.

    """
    def __init__(self, sensors, select=[], proj='default', size=1,
                 color='k', marker='.', frame=0.05, **kwargs):
        sensors = as_sensor(sensors)

        # layout figure
        self._drag_ax = None
        self._drag_x = None
        self._drag_y = None
        layout = ImLayout(4, 1, 3, None, {}, frame='off', ncol=2, nrow=2, **kwargs)
        EelFigure.__init__(self, sensors.sysname, layout)
        self.figure.subplots_adjust(left=0, bottom=0, right=1, top=1,
                                    wspace=.1, hspace=.1)

        # store args
        self._sensors = sensors

        min_coords = sensors.locs.min(0) - frame
        max_coords = sensors.locs.max(0) + frame
        xlim, ylim, zlim = list(zip(min_coords, max_coords))

        # back
        ax = self.ax0 = self.axes[0]
        ax.proj = 'back'  # y-
        ax.extent = False
        ax.set_xlim(xlim)
        ax.set_ylim(zlim)
        self._h0 = _ax_map2d(ax, sensors, ax.proj, ax.extent, size, color, marker)

        # left
        ax = self.ax1 = self.axes[1]
        ax.proj = 'left'  # x-
        ax.extent = False
        ax.set_xlim(ylim)
        ax.set_ylim(zlim)
        self._h1 = _ax_map2d(ax, sensors, ax.proj, ax.extent, size, color, marker)

        # top
        ax = self.ax2 = self.axes[2]
        ax.proj = 'top'  # z+
        ax.extent = False
        ax.set_xlim(xlim)
        ax.set_ylim(ylim)
        self._h2 = _ax_map2d(ax, sensors, ax.proj, ax.extent, size, color, marker)

        # proj
        ax = self.ax3 = self.axes[3]
        ax.proj = proj
        ax.extent = 1
        self._h3 = _ax_map2d(ax, sensors, ax.proj, ax.extent, size, color, marker)
        self.ax3.set_xlim(-frame, 1 + frame)
        self.ax3.set_ylim(-frame, 1 + frame)

        self._sensor_maps = (self._h0, self._h1, self._h2, self._h3)

        # selection
        self.sel_kwargs = dict(marker='o', s=5, c='r', linewidths=.9)
        self._sel_h = []
        if select is not None:
            self.set_selection(select)
        else:
            self._selection = None

        # setup mpl event handling
        self.canvas.mpl_connect("button_press_event", self._on_button_press)
        self.canvas.mpl_connect("button_release_event", self._on_button_release)

        self._show()

    def _fill_toolbar(self, tb):
        from .._wxgui import wx

        tb.AddSeparator()

        # plot labels
        btn = wx.Button(tb, wx.ID_CLEAR, "Clear")
        tb.AddControl(btn)
        btn.Bind(wx.EVT_BUTTON, self._OnClear)

    def clear(self):
        "Clear the current sensor selection."
        self._selection = None
        self._update_mark_plot()

    def get_selection(self):
        """Retrieve the current selection

        Returns
        -------
        selection : list
            The current selection as a list of indices.
        """
        if self._selection is None:
            return []
        else:
            return np.where(self._selection)[0]

    def _on_button_press(self, event):
        ax = event.inaxes
        if ax:
            self._is_dragging = True
            self._drag_ax = event.inaxes
            self._drag_x = event.xdata
            self._drag_y = event.ydata

            self._frame.store_canvas()
            x = np.ones(5) * event.xdata
            y = np.ones(5) * event.ydata
            self._drag_rect = ax.plot(x, y, '-k')[0]

    def _on_button_release(self, event):
        if not hasattr(self, '_drag_rect'):
            return

        x = self._drag_rect.get_xdata()
        y = self._drag_rect.get_ydata()
        xmin = min(x)
        xmax = max(x)
        ymin = min(y)
        ymax = max(y)

        ax = self._drag_ax
        locs = self._sensors.get_locs_2d(ax.proj, ax.extent, SENSORMAP_FRAME)
        x = locs[:, 0]
        y = locs[:, 1]
        sel = (x > xmin) & (x < xmax) & (y > ymin) & (y < ymax)

        if self._selection is None:
            self._selection = sel
        elif event.button == 1:
            self._selection[sel] = True
        else:
            self._selection[sel] = False

        # clear dragging-related attributes
        self._drag_rect.remove()
        del self._drag_rect
        self._drag_ax = None
        self._drag_x = None
        self._drag_y = None

        self._update_mark_plot()

    @staticmethod
    def _on_motion_status_text(event):
        ax = event.inaxes
        if ax:
            out = "Projection: " + ax.proj
            if not ax.extent:
                out += ';  x = %i, y = %i' % (round(event.xdata * 1e3),
                                              round(event.ydata * 1e3))
            return out
        return ''

    def _on_motion_sub(self, event):
        ax = event.inaxes
        if ax and ax is self._drag_ax:
            x0 = self._drag_x
            x1 = event.xdata
            y0 = self._drag_y
            y1 = event.ydata
            x = [x0, x1, x1, x0, x0]
            y = [y0, y0, y1, y1, y0]
            self._drag_rect.set_data(x, y)
            return {ax}
        return set()

    def _OnClear(self, event):
        self.clear()

    def set_selection(self, select):
        """
        Set the current selection with a list of indices.

        Parameters
        ----------
        select : sensor index
            Index for sensor dimension, for example array_like of int, or list
            of sensor names.
        """
        idx = self._sensors._array_index(select)
        self._selection = np.zeros(len(self._sensors), dtype=bool)
        self._selection[idx] = True
        self._update_mark_plot()

    def _update_mark_plot(self):
        for h in self._sensor_maps:
            h.sensors.mark_sensors(None)  # clear old selection
            h.sensors.mark_sensors(self._selection, **self.sel_kwargs)
        self.canvas.draw()


class SensorMap(SensorMapMixin, EelFigure):
    """Plot sensor positions in 2 dimensions

    Parameters
    ----------
    sensors : Sensor | NDVar
        The :class:`Sensor` dimension, or an :class:`NDVar` with a sensor
        dimension.
    labels : 'none' | 'index' | 'name' | 'fullname'
        Content of the labels. For 'name', any prefix common to all names
        is removed; with 'fullname', the full name is shown.
    proj:
        Transform to apply to 3 dimensional sensor coordinates for plotting
        locations in a plane
    size : scalar
        Size for the sensor markers.
    color : matplotlib color
        Color for the sensor markers.
    marker : str
        Marker for the sensor positions.
    mark : None | list of int
        List of sensor indices to mark.
    head_radius : scalar | tuple | True
        Radius of the head outline drawn over sensors (on sensor plots with
        normalized positions, 0.45 is the outline of the topomap); 0 to plot no
        outline; tuple for separate (right, anterior) radius. True to be equal
        to :class:`plot.Topomap` with ``method="mne"``.
        The default is determined automatically.
    head_pos : scalar
        Head outline position along the anterior axis (0 is the center, 0.5 is
        the top end of the plot).
    connectivity : bool
        Show sensor connectivity (default False).
    ...
        Also accepts :ref:`general-layout-parameters`.
    """
    def __init__(self, sensors, labels='name', proj='default', size=1,
                 color='k', marker='.', mark=None, head_radius=None,
                 head_pos=0., connectivity=False, margins=None, **kwargs):
        sensors = as_sensor(sensors)
        kwargs.setdefault('frame', 'none')
        layout = ImLayout(1, 1, 5, margins, {}, **kwargs)
        EelFigure.__init__(self, sensors.sysname, layout)
        ax = self.axes[0]

        # store args
        self._sensors = sensors
        self._proj = proj
        self._marker_handles = []
        self._connectivity = None

        self._markers = _ax_map2d(ax, sensors, proj, 1, size, color, marker,
                                  mark, head_radius, head_pos, labels=labels)
        SensorMapMixin.__init__(self, [self._markers.sensors])

        if connectivity:
            self.show_connectivity()

        self._show()

    def remove_markers(self):
        "Remove all sensor markers."
        while len(self._marker_handles) > 0:
            h = self._marker_handles.pop(0)
            h.remove()
        self.canvas.draw()

    def show_connectivity(self, show=True):
        """Show the sensor connectivity as lines connecting sensors.

        Parameters
        ----------
        show : bool
            Show or hide the sensor connectivity.
        """
        if show:
            conn = self._sensors.connectivity()
            self._markers.connectivity.show(conn)
        else:
            self._markers.connectivity.show(None)
        self.draw()


def map3d(sensors, marker='c*', labels=False, head=0):
    """3d plot of a Sensors instance"""
    import matplotlib.pyplot as plt

    sensors = as_sensor(sensors)

    locs = sensors.locs
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    ax.scatter(locs[:, 0], locs[:, 1], locs[:, 2])
    # plot head ball
    if head:
        u = np.linspace(0, 1 * np.pi, 10)
        v = np.linspace(0, np.pi, 10)

        x = 5 * head * np.outer(np.cos(u), np.sin(v))
        z = 10 * (head * np.outer(np.sin(u), np.sin(v)) - .5)  # vertical
        y = 5 * head * np.outer(np.ones(np.size(u)), np.cos(v))  # axis of the sphere
        ax.plot_surface(x, y, z, rstride=1, cstride=1, color='w')
    # n = 100
    # for c, zl, zh in [('r', -50, -25), ('b', -30, -5)]:
    # xs, ys, zs = zip(*
    #               [(random.randrange(23, 32),
    #                 random.randrange(100),
    #                 random.randrange(zl, zh)
    #                 ) for i in range(n)])
    # ax.scatter(xs, ys, zs, c=c)



import os
import itertools
import pytest
from eelbrain.testing import TempDir

from ..stc_dataset import DatasetSTCLoader


def _create_fake_files(tmpdir):
    """Write empty stc files to tmpdir to be detected
    by the DatasetSTCLoader"""
    subjects = ("R0000", "A9999")
    factor1 = ("level-a", "level-b")
    factor2 = ("noun", "verb")
    folder_combos = itertools.product(factor1, factor2)
    file_combos = itertools.product(subjects, factor1, factor2)
    folders = ["_".join(f) for f in folder_combos]
    files = ["_".join(f) + "-lh.stc" for f in file_combos]
    for f in folders:
        folder_path = os.path.join(tmpdir, f)
        if not os.path.exists(folder_path):
            os.mkdir(folder_path)
        files_to_write = filter(lambda x: f in x, files)
        for fw in files_to_write:
            file_path = os.path.join(folder_path, fw)
            open(file_path, "a").close()


def test_loader():
    tmp = TempDir()
    _create_fake_files(tmp)
    loader = DatasetSTCLoader(tmp)
    levels = loader.levels
    # check detected level names
    assert "level-a" in levels[0]
    assert "level-b" in levels[0]
    assert "noun" in levels[1]
    assert "verb" in levels[1]
    assert "R0000" in loader.subjects
    with pytest.raises(ValueError):
        loader.set_factor_names(["only-one-factor"])
    loader.set_factor_names(["factor1", "factor2"])
    assert loader.design_shape == "2 x 2"
    ds = loader.make_dataset(load_stcs=False)
    assert "factor1" in ds
    assert "factor2" in ds
    assert "subject" in ds
    assert ds["subject"].random



#  Utilities working with (and needing to import) plots
import os

import numpy as np

from .._utils import ui
from ._base import TimeSlicer


def save_movie(figures, filename=None, time_dilation=4, tstart=None, tstop=None, size=None, **kwargs):
    """Save a movie combining multiple figures with moving time axes

    Parameters
    ----------
    figures : dict
        ``{(xpos, ypos): figure}`` dictionary indicating placement of figures.
    filename : str
        Filename for the movie (omit to use a GUI).
    time_dilation : float
        Factor by which to stretch time (default 4). Time dilation is
        controlled through the frame-rate; if the ``fps`` keyword argument
        is specified, ``time_dilation`` is ignored.
    tstart : float
        Time axis start (default is earliest time in ``figures``).
    tstop : float
        Time axis stop (default includes latest time in ``figures``).
    ...
        :func:`imageio.mimwrite` parmeters.
    """
    import imageio
    from PIL import Image

    if filename is None:
        filename = ui.ask_saveas("Save movie...", None, [('Movie (*.mov)', '*.mov')])
        if not filename:
            return
    else:
        filename = os.path.expanduser(filename)

    time_dims = list(filter(None, (getattr(f, '_time_dim', None) for f in figures.values())))
    if tstart is None:
        tstart = min(dim.tmin for dim in time_dims)
    if tstop is None:
        tstop = max(dim.tstop for dim in time_dims)
    if 'fps' in kwargs:
        tstep = time_dilation / kwargs['fps']
    else:
        tstep = min(dim.tstep for dim in time_dims)
        kwargs['fps'] = 1. / tstep / time_dilation

    times = np.arange(tstart, tstop, tstep)

    ims = []
    x_max = y_max = None
    for t in times:
        t_ims = []
        for (x, y), fig in figures.items():
            fig.set_time(t)
            f_im = Image.fromarray(fig._im_array(), 'RGBA')
            t_ims.append((x, y, f_im))

        if x_max is None:
            x_max = max(x + im.size[0] for x, y, im in t_ims)
            y_max = max(y + im.size[1] for x, y, im in t_ims)
        im_buf = Image.new('RGBA', (x_max, y_max), (1, 1, 1, 0))

        for x, y, im in t_ims:
            im_buf.paste(im, (x, y))

        ims.append(np.array(im_buf))
    imageio.mimwrite(filename, ims, **kwargs)
